from unittest import (
    TestCase,
    mock,
)

from lxml import etree

from pcs import settings
from pcs.common import reports
from pcs.common.fencing_topology import (
    TARGET_TYPE_ATTRIBUTE,
    TARGET_TYPE_NODE,
    TARGET_TYPE_REGEXP,
)
from pcs.common.reports import ReportItemSeverity as severity
from pcs.common.reports import codes as report_codes
from pcs.common.reports.item import ReportItem
from pcs.lib.cib import fencing_topology as lib
from pcs.lib.errors import LibraryError
from pcs.lib.pacemaker.state import ClusterState

from pcs_test.tools import (
    fixture,
    fixture_crm_mon,
)
from pcs_test.tools.assertions import (
    assert_report_item_list_equal,
    assert_xml_equal,
)
from pcs_test.tools.custom_mock import MockLibraryReportProcessor
from pcs_test.tools.misc import create_patcher
from pcs_test.tools.misc import get_test_resource as rc
from pcs_test.tools.xml import etree_to_str

patch_lib = create_patcher("pcs.lib.cib.fencing_topology")

# pylint: disable=too-few-public-methods
# pylint: disable=protected-access
# pylint: disable=no-self-use


class CibMixin:
    @staticmethod
    def get_cib():
        return etree.fromstring(
            """
            <cib><configuration>
            <fencing-topology>
                <fencing-level
                    id="fl1" index="1" devices="d1,d2" target="nodeA"
                />
                <fencing-level
                    id="fl2" index="2" devices="d3" target="nodeA"
                />
                <fencing-level
                    id="fl3" index="1" devices="d2,d1" target="nodeB"
                />
                <fencing-level
                    id="fl4" index="2" devices="d3" target="nodeB"
                />
                <fencing-level
                    id="fl5" index="1" devices="d3,d4" target-pattern="node\\d+"
                />
                <fencing-level
                    id="fl6" index="2" devices="d1" target-pattern="node\\d+"
                />
                <fencing-level
                    id="fl7" index="3" devices="d3,d4"
                    target-attribute="fencing" target-value="improved"
                />
                <fencing-level
                    id="fl8" index="4" devices="d5"
                    target-attribute="fencing" target-value="improved"
                />
                <fencing-level
                    id="fl9" index="3" devices="dR" target-pattern="node-R.*"
                />
                <fencing-level
                    id="fl10" index="4" devices="dR-special"
                    target-attribute="fencing" target-value="remote-special"
                />
            </fencing-topology>
            </configuration></cib>
        """
        )


class StatusNodesMixin:
    def get_status(self):
        with open(rc("crm_mon.minimal.xml")) as crm_mon_file:
            crm_mon_xml = crm_mon_file.read()
        return ClusterState(
            fixture_crm_mon.complete_state(
                crm_mon_xml,
                nodes_xml="""
                <nodes>
                    <node name="nodeA" id="1" is_dc="true" />
                    <node name="nodeB" id="2" />
                </nodes>
            """,
            )
        ).node_section.nodes


@patch_lib("_append_level_element")
@patch_lib("_validate_level_target_devices_does_not_exist", return_value=[])
@patch_lib("_validate_devices", return_value=[])
@patch_lib("_validate_target", return_value=[])
@patch_lib("_validate_level", return_value=([], "valid_level"))
class AddLevel(TestCase):
    # pylint: disable=too-many-instance-attributes
    def setUp(self):
        self.reporter = MockLibraryReportProcessor()
        self.topology_el = "a topology element"
        self.resources_el = "a resources element"
        self.level = "a level"
        self.target_type = "a target type"
        self.target_value = "a target value"
        self.devices = ["device1", "device2"]
        self.cluster_status_nodes = "a status"
        self.force_device = "a force for a device"
        self.force_node = "a force for a node"

    def assert_validators_called(
        self,
        mock_val_level,
        mock_val_target,
        mock_val_devices,
        mock_val_dupl,
        dupl_called=True,
    ):
        mock_val_level.assert_called_once_with(self.level)
        mock_val_target.assert_called_once_with(
            self.cluster_status_nodes,
            self.target_type,
            self.target_value,
            self.force_node,
        )
        mock_val_devices.assert_called_once_with(
            self.resources_el, self.devices, self.force_device
        )
        if dupl_called:
            mock_val_dupl.assert_called_once_with(
                self.topology_el,
                self.level,
                self.target_type,
                self.target_value,
                self.devices,
            )
        else:
            mock_val_dupl.assert_not_called()

    def assert_called_invalid(
        self,
        mock_val_level,
        mock_val_target,
        mock_val_devices,
        mock_val_dupl,
        mock_append,
        dupl_called=True,
        report_list=None,
    ):
        report_list = report_list or []
        with self.assertRaises(LibraryError):
            lib.add_level(
                self.reporter,
                self.topology_el,
                self.resources_el,
                self.level,
                self.target_type,
                self.target_value,
                self.devices,
                self.cluster_status_nodes,
                self.force_device,
                self.force_node,
            )
        assert_report_item_list_equal(
            self.reporter.report_item_list, report_list
        )
        self.assert_validators_called(
            mock_val_level,
            mock_val_target,
            mock_val_devices,
            mock_val_dupl,
            dupl_called,
        )
        mock_append.assert_not_called()

    def test_success(
        self,
        mock_val_level,
        mock_val_target,
        mock_val_devices,
        mock_val_dupl,
        mock_append,
    ):
        lib.add_level(
            self.reporter,
            self.topology_el,
            self.resources_el,
            self.level,
            self.target_type,
            self.target_value,
            self.devices,
            self.cluster_status_nodes,
            self.force_device,
            self.force_node,
        )
        self.assert_validators_called(
            mock_val_level, mock_val_target, mock_val_devices, mock_val_dupl
        )
        mock_append.assert_called_once_with(
            self.topology_el,
            "valid_level",
            self.target_type,
            self.target_value,
            self.devices,
        )

    def test_invalid_level(
        self,
        mock_val_level,
        mock_val_target,
        mock_val_devices,
        mock_val_dupl,
        mock_append,
    ):
        mock_val_level.return_value = (
            [
                reports.item.ReportItem.error(
                    reports.messages.InvalidOptionValue(
                        "level", self.level, "a positive integer"
                    )
                )
            ],
            None,
        )
        report_list = [
            fixture.error(
                report_codes.INVALID_OPTION_VALUE,
                option_value=self.level,
                option_name="level",
                allowed_values="a positive integer",
                cannot_be_empty=False,
                forbidden_characters=None,
            ),
        ]
        self.assert_called_invalid(
            mock_val_level,
            mock_val_target,
            mock_val_devices,
            mock_val_dupl,
            mock_append,
            dupl_called=False,
            report_list=report_list,
        )

    def test_invalid_target(
        self,
        mock_val_level,
        mock_val_target,
        mock_val_devices,
        mock_val_dupl,
        mock_append,
    ):
        mock_val_target.return_value = [
            ReportItem.error(reports.messages.NodeNotFound(self.target_value))
        ]
        report_list = [
            fixture.error(
                report_codes.NODE_NOT_FOUND,
                node=self.target_value,
                searched_types=[],
            ),
        ]
        self.assert_called_invalid(
            mock_val_level,
            mock_val_target,
            mock_val_devices,
            mock_val_dupl,
            mock_append,
            dupl_called=False,
            report_list=report_list,
        )

    def test_invalid_devices(
        self,
        mock_val_level,
        mock_val_target,
        mock_val_devices,
        mock_val_dupl,
        mock_append,
    ):
        mock_val_devices.return_value = [
            ReportItem.error(
                reports.messages.StonithResourcesDoNotExist(self.devices)
            )
        ]
        report_list = [
            fixture.error(
                report_codes.STONITH_RESOURCES_DO_NOT_EXIST,
                stonith_ids=self.devices,
            ),
        ]
        self.assert_called_invalid(
            mock_val_level,
            mock_val_target,
            mock_val_devices,
            mock_val_dupl,
            mock_append,
            dupl_called=False,
            report_list=report_list,
        )

    def test_already_exists(
        self,
        mock_val_level,
        mock_val_target,
        mock_val_devices,
        mock_val_dupl,
        mock_append,
    ):
        mock_val_dupl.return_value = [
            reports.item.ReportItem.error(
                reports.messages.CibFencingLevelAlreadyExists(
                    self.level,
                    self.target_type,
                    self.target_value,
                    self.devices,
                )
            )
        ]
        report_list = [
            fixture.error(
                report_codes.CIB_FENCING_LEVEL_ALREADY_EXISTS,
                level=self.level,
                target_type=self.target_type,
                target_value=self.target_value,
                devices=self.devices,
            ),
        ]
        self.assert_called_invalid(
            mock_val_level,
            mock_val_target,
            mock_val_devices,
            mock_val_dupl,
            mock_append,
            dupl_called=True,
            report_list=report_list,
        )


class RemoveAllLevels(TestCase, CibMixin):
    def setUp(self):
        self.cib = self.get_cib()
        self.tree = self.cib.find("configuration/fencing-topology")

    def test_success(self):
        lib.remove_all_levels(self.tree)
        assert_xml_equal("<fencing-topology />", etree_to_str(self.tree))


class RemoveLevelsByParams(TestCase, CibMixin):
    def setUp(self):
        self.cib = self.get_cib()
        self.tree = self.cib.find("configuration/fencing-topology")
        self.reporter = MockLibraryReportProcessor()

    def get_remaining_ids(self):
        self.assertTrue(
            self.cib.find("configuration/fencing-topology") is not None
        )
        return [el.get("id") for el in self.tree.findall("fencing-level")]

    def test_level(self):
        report_list = lib.remove_levels_by_params(self.tree, level=2)
        self.assertEqual(
            self.get_remaining_ids(),
            ["fl1", "fl3", "fl5", "fl7", "fl8", "fl9", "fl10"],
        )
        assert_report_item_list_equal(report_list, [])

    def test_target_node(self):
        report_list = lib.remove_levels_by_params(
            self.tree, target_type=TARGET_TYPE_NODE, target_value="nodeA"
        )
        self.assertEqual(
            self.get_remaining_ids(),
            ["fl3", "fl4", "fl5", "fl6", "fl7", "fl8", "fl9", "fl10"],
        )
        assert_report_item_list_equal(report_list, [])

    def test_target_pattern(self):
        report_list = lib.remove_levels_by_params(
            self.tree, target_type=TARGET_TYPE_REGEXP, target_value=r"node\d+"
        )
        self.assertEqual(
            self.get_remaining_ids(),
            ["fl1", "fl2", "fl3", "fl4", "fl7", "fl8", "fl9", "fl10"],
        )
        assert_report_item_list_equal(report_list, [])

    def test_target_attrib(self):
        report_list = lib.remove_levels_by_params(
            self.tree,
            target_type=TARGET_TYPE_ATTRIBUTE,
            target_value=("fencing", "improved"),
        )
        self.assertEqual(
            self.get_remaining_ids(),
            ["fl1", "fl2", "fl3", "fl4", "fl5", "fl6", "fl9", "fl10"],
        )
        assert_report_item_list_equal(report_list, [])

    def test_one_device(self):
        report_list = lib.remove_levels_by_params(self.tree, devices=["d3"])
        self.assertEqual(
            self.get_remaining_ids(),
            ["fl1", "fl3", "fl5", "fl6", "fl7", "fl8", "fl9", "fl10"],
        )
        assert_report_item_list_equal(report_list, [])

    def test_more_devices(self):
        report_list = lib.remove_levels_by_params(
            self.tree, devices=["d2", "d1"]
        )
        self.assertEqual(
            self.get_remaining_ids(),
            ["fl1", "fl2", "fl4", "fl5", "fl6", "fl7", "fl8", "fl9", "fl10"],
        )
        assert_report_item_list_equal(report_list, [])

    def test_combination(self):
        report_list = lib.remove_levels_by_params(
            self.tree, 2, TARGET_TYPE_NODE, "nodeB", ["d3"]
        )
        self.assertEqual(
            self.get_remaining_ids(),
            ["fl1", "fl2", "fl3", "fl5", "fl6", "fl7", "fl8", "fl9", "fl10"],
        )
        assert_report_item_list_equal(report_list, [])

    def test_invalid_target(self):
        report_list = lib.remove_levels_by_params(
            self.tree, target_type="bad_target", target_value="nodeA"
        )
        assert_report_item_list_equal(
            report_list,
            [
                (
                    severity.ERROR,
                    report_codes.INVALID_OPTION_TYPE,
                    {
                        "option_name": "target",
                        "allowed_types": [
                            "node",
                            "regular expression",
                            "attribute_name=value",
                        ],
                    },
                    None,
                ),
            ],
        )
        self.assertEqual(
            self.get_remaining_ids(),
            [
                "fl1",
                "fl2",
                "fl3",
                "fl4",
                "fl5",
                "fl6",
                "fl7",
                "fl8",
                "fl9",
                "fl10",
            ],
        )

    def test_no_such_level(self):
        report_list = lib.remove_levels_by_params(
            self.tree, 9, TARGET_TYPE_NODE, "nodeB", ["d3"]
        )
        assert_report_item_list_equal(
            report_list,
            [
                (
                    severity.ERROR,
                    report_codes.CIB_FENCING_LEVEL_DOES_NOT_EXIST,
                    {
                        "devices": [
                            "d3",
                        ],
                        "target_type": TARGET_TYPE_NODE,
                        "target_value": "nodeB",
                        "level": 9,
                    },
                    None,
                ),
            ],
        )
        self.assertEqual(
            self.get_remaining_ids(),
            [
                "fl1",
                "fl2",
                "fl3",
                "fl4",
                "fl5",
                "fl6",
                "fl7",
                "fl8",
                "fl9",
                "fl10",
            ],
        )

    def test_no_such_level_ignore_missing(self):
        report_list = lib.remove_levels_by_params(
            self.tree, 9, TARGET_TYPE_NODE, "nodeB", ["d3"], True
        )
        assert_report_item_list_equal(report_list, [])
        self.assertEqual(
            self.get_remaining_ids(),
            [
                "fl1",
                "fl2",
                "fl3",
                "fl4",
                "fl5",
                "fl6",
                "fl7",
                "fl8",
                "fl9",
                "fl10",
            ],
        )


class RemoveDeviceFromAllLevels(TestCase, CibMixin):
    def setUp(self):
        self.cib = self.get_cib()
        self.tree = self.cib.find("configuration/fencing-topology")

    def test_success(self):
        lib.remove_device_from_all_levels(self.tree, "d3")
        assert_xml_equal(
            """
            <fencing-topology>
                <fencing-level
                    id="fl1" index="1" devices="d1,d2" target="nodeA"
                />
                <fencing-level
                    id="fl3" index="1" devices="d2,d1" target="nodeB"
                />
                <fencing-level
                    id="fl5" index="1" devices="d4" target-pattern="node\\d+"
                />
                <fencing-level
                    id="fl6" index="2" devices="d1" target-pattern="node\\d+"
                />
                <fencing-level
                    id="fl7" index="3" devices="d4"
                    target-attribute="fencing" target-value="improved"
                />
                <fencing-level
                    id="fl8" index="4" devices="d5"
                    target-attribute="fencing" target-value="improved"
                />
                <fencing-level
                    id="fl9" index="3" devices="dR" target-pattern="node-R.*"
                />
                <fencing-level
                    id="fl10" index="4" devices="dR-special"
                    target-attribute="fencing" target-value="remote-special"
                />
            </fencing-topology>
            """,
            etree_to_str(self.tree),
        )

    def test_no_such_device(self):
        original_xml = etree_to_str(self.tree)
        lib.remove_device_from_all_levels(self.tree, "dX")
        assert_xml_equal(original_xml, etree_to_str(self.tree))


class Export(TestCase, CibMixin):
    def setUp(self):
        self.cib = self.get_cib()
        self.tree = self.cib.find("configuration/fencing-topology")

    def test_empty(self):
        self.assertEqual(
            lib.export(etree.fromstring("<fencing-topology />")), []
        )

    def test_success(self):
        self.assertEqual(
            lib.export(self.tree),
            [
                {
                    "level": "1",
                    "target_type": "node",
                    "target_value": "nodeA",
                    "devices": ["d1", "d2"],
                },
                {
                    "level": "2",
                    "target_type": "node",
                    "target_value": "nodeA",
                    "devices": ["d3"],
                },
                {
                    "level": "1",
                    "target_type": "node",
                    "target_value": "nodeB",
                    "devices": ["d2", "d1"],
                },
                {
                    "level": "2",
                    "target_type": "node",
                    "target_value": "nodeB",
                    "devices": ["d3"],
                },
                {
                    "level": "1",
                    "target_type": "regexp",
                    "target_value": r"node\d+",
                    "devices": ["d3", "d4"],
                },
                {
                    "level": "2",
                    "target_type": "regexp",
                    "target_value": r"node\d+",
                    "devices": ["d1"],
                },
                {
                    "level": "3",
                    "target_type": "attribute",
                    "target_value": ("fencing", "improved"),
                    "devices": ["d3", "d4"],
                },
                {
                    "level": "4",
                    "target_type": "attribute",
                    "target_value": ("fencing", "improved"),
                    "devices": ["d5"],
                },
                {
                    "level": "3",
                    "target_type": "regexp",
                    "target_value": "node-R.*",
                    "devices": ["dR"],
                },
                {
                    "level": "4",
                    "target_type": "attribute",
                    "target_value": ("fencing", "remote-special"),
                    "devices": ["dR-special"],
                },
            ],
        )


class Verify(TestCase, CibMixin, StatusNodesMixin):
    def setUp(self):
        self.cib = self.get_cib()
        self.tree = self.cib.find("configuration/fencing-topology")

    def fixture_resource(self, tree, name):
        el = etree.SubElement(tree, "primitive", id=name, type="fence_dummy")
        el.set("class", "stonith")

    @mock.patch.object(
        settings,
        "pacemaker_api_result_schema",
        rc("pcmk_api_rng/api-result.rng"),
    )
    def test_empty(self):
        resources = etree.fromstring("<resources />")
        topology = etree.fromstring("<fencing-topology />")

        report_list = lib.verify(topology, resources, self.get_status())

        assert_report_item_list_equal(report_list, [])

    @mock.patch.object(
        settings,
        "pacemaker_api_result_schema",
        rc("pcmk_api_rng/api-result.rng"),
    )
    def test_success(self):
        resources = etree.fromstring("<resources />")
        for name in ["d1", "d2", "d3", "d4", "d5", "dR", "dR-special"]:
            self.fixture_resource(resources, name)

        report_list = lib.verify(self.tree, resources, self.get_status())

        assert_report_item_list_equal(report_list, [])

    def test_failures(self):
        resources = etree.fromstring("<resources />")

        report_list = lib.verify(self.tree, resources, [])

        report = [
            (
                severity.ERROR,
                report_codes.STONITH_RESOURCES_DO_NOT_EXIST,
                {
                    "stonith_ids": [
                        "d1",
                        "d2",
                        "d3",
                        "d4",
                        "d5",
                        "dR",
                        "dR-special",
                    ],
                },
                None,
            ),
            (
                severity.ERROR,
                report_codes.NODE_NOT_FOUND,
                {
                    "node": "nodeA",
                    "searched_types": [],
                },
                None,
            ),
            (
                severity.ERROR,
                report_codes.NODE_NOT_FOUND,
                {
                    "node": "nodeB",
                    "searched_types": [],
                },
                None,
            ),
        ]
        assert_report_item_list_equal(report_list, report)


class ValidateLevel(TestCase):
    def test_success(self):
        level_list = [
            (1, 1),
            ("1", 1),
            (9, 9),
            ("9", 9),
            ("05", 5),
        ]
        for level_in, level_out in level_list:
            with self.subTest(level=level_in):
                report_list, valid_level = lib._validate_level(level_in)
                self.assertEqual(level_out, valid_level)
                assert_report_item_list_equal(report_list, [])

    def test_invalid(self):
        level_list = ["", 0, "0", -1, "-1", "1abc"]
        for level in level_list:
            with self.subTest(level=level):
                report_list, valid_level = lib._validate_level(level)
                self.assertEqual(None, valid_level)
                assert_report_item_list_equal(
                    report_list,
                    [
                        fixture.error(
                            report_codes.INVALID_OPTION_VALUE,
                            option_value=level,
                            option_name="level",
                            allowed_values="a positive integer",
                            cannot_be_empty=False,
                            forbidden_characters=None,
                        ),
                    ],
                )


@patch_lib("_validate_target_valuewise")
@patch_lib("_validate_target_typewise")
class ValidateTarget(TestCase):
    def test_delegate(self, validate_type, validate_value):
        lib._validate_target("status", "type", "value", "force")
        validate_type.assert_called_once_with("type")
        validate_value.assert_called_once_with(
            "status", "type", "value", "force"
        )


class ValidateTargetTypewise(TestCase):
    def test_success(self):
        report_list = []
        report_list.extend(lib._validate_target_typewise(TARGET_TYPE_NODE))
        report_list.extend(lib._validate_target_typewise(TARGET_TYPE_ATTRIBUTE))
        report_list.extend(lib._validate_target_typewise(TARGET_TYPE_REGEXP))
        assert_report_item_list_equal(report_list, [])

    def test_empty(self):
        report_list = lib._validate_target_typewise("")
        report = [
            (
                severity.ERROR,
                report_codes.INVALID_OPTION_TYPE,
                {
                    "option_name": "target",
                    "allowed_types": [
                        "node",
                        "regular expression",
                        "attribute_name=value",
                    ],
                },
                None,
            )
        ]
        assert_report_item_list_equal(report_list, report)

    def test_invalid(self):
        report_list = lib._validate_target_typewise("bad_target")
        report = [
            (
                severity.ERROR,
                report_codes.INVALID_OPTION_TYPE,
                {
                    "option_name": "target",
                    "allowed_types": [
                        "node",
                        "regular expression",
                        "attribute_name=value",
                    ],
                },
                None,
            )
        ]
        assert_report_item_list_equal(report_list, report)


@mock.patch.object(
    settings, "pacemaker_api_result_schema", rc("pcmk_api_rng/api-result.rng")
)
class ValidateTargetValuewise(TestCase, StatusNodesMixin):
    def test_node_valid(self):
        state = self.get_status()
        assert_report_item_list_equal(
            lib._validate_target_valuewise(state, TARGET_TYPE_NODE, "nodeA"),
            [],
        )

    def test_node_empty(self):
        state = self.get_status()
        report_list = lib._validate_target_valuewise(
            state, TARGET_TYPE_NODE, ""
        )
        report = [
            (
                severity.ERROR,
                report_codes.NODE_NOT_FOUND,
                {
                    "node": "",
                    "searched_types": [],
                },
                report_codes.FORCE,
            )
        ]
        assert_report_item_list_equal(report_list, report)

    def test_node_invalid(self):
        state = self.get_status()
        report_list = lib._validate_target_valuewise(
            state, TARGET_TYPE_NODE, "rh7-x"
        )
        report = [
            (
                severity.ERROR,
                report_codes.NODE_NOT_FOUND,
                {
                    "node": "rh7-x",
                    "searched_types": [],
                },
                report_codes.FORCE,
            )
        ]
        assert_report_item_list_equal(report_list, report)

    def test_node_invalid_force(self):
        state = self.get_status()
        report_list = lib._validate_target_valuewise(
            state, TARGET_TYPE_NODE, "rh7-x", force_node=True
        )
        report = [
            (
                severity.WARNING,
                report_codes.NODE_NOT_FOUND,
                {
                    "node": "rh7-x",
                    "searched_types": [],
                },
                None,
            )
        ]
        assert_report_item_list_equal(report_list, report)

    def test_node_invalid_not_forceable(self):
        state = self.get_status()
        report_list = lib._validate_target_valuewise(
            state, TARGET_TYPE_NODE, "rh7-x", allow_force=False
        )
        report = [
            (
                severity.ERROR,
                report_codes.NODE_NOT_FOUND,
                {
                    "node": "rh7-x",
                    "searched_types": [],
                },
                None,
            )
        ]
        assert_report_item_list_equal(report_list, report)


class ValidateDevices(TestCase):
    def setUp(self):
        self.resources_el = etree.fromstring(
            """
            <resources>
                <primitive id="dummy"
                    class="ocf" provider="pacemaker" type="Stateful"
                 />
                <primitive id="stonith1" class="stonith" type="fence_xvm" />
                <primitive id="stonith2" class="stonith" type="fence_apc" />
            </resources>
        """
        )

    def test_success(self):
        report_list = []
        report_list.extend(
            lib._validate_devices(self.resources_el, ["stonith1"])
        )
        report_list.extend(
            lib._validate_devices(self.resources_el, ["stonith1", "stonith2"])
        )
        assert_report_item_list_equal(report_list, [])

    def test_empty(self):
        report_list = lib._validate_devices(self.resources_el, [])
        report = [
            (
                severity.ERROR,
                report_codes.REQUIRED_OPTIONS_ARE_MISSING,
                {
                    "option_type": None,
                    "option_names": ["stonith devices"],
                },
                None,
            )
        ]
        assert_report_item_list_equal(report_list, report)

    def test_invalid(self):
        report_list = lib._validate_devices(
            self.resources_el, ["dummy", "fenceX"]
        )
        report = [
            (
                severity.ERROR,
                report_codes.STONITH_RESOURCES_DO_NOT_EXIST,
                {
                    "stonith_ids": ["dummy", "fenceX"],
                },
                report_codes.FORCE,
            )
        ]
        assert_report_item_list_equal(report_list, report)

    def test_invalid_forced(self):
        report_list = lib._validate_devices(
            self.resources_el, ["dummy", "fenceX"], force_device=True
        )
        report = [
            (
                severity.WARNING,
                report_codes.STONITH_RESOURCES_DO_NOT_EXIST,
                {
                    "stonith_ids": ["dummy", "fenceX"],
                },
                None,
            )
        ]
        assert_report_item_list_equal(report_list, report)

    def test_node_invalid_not_forceable(self):
        report_list = lib._validate_devices(
            self.resources_el, ["dummy", "fenceX"], allow_force=False
        )
        report = [
            (
                severity.ERROR,
                report_codes.STONITH_RESOURCES_DO_NOT_EXIST,
                {
                    "stonith_ids": ["dummy", "fenceX"],
                },
                None,
            )
        ]
        assert_report_item_list_equal(report_list, report)


@patch_lib("_find_level_elements")
class ValidateLevelTargetDevicesDoesNotExist(TestCase):
    def test_success(self, mock_find):
        mock_find.return_value = []

        report_list = lib._validate_level_target_devices_does_not_exist(
            "tree", "level", "target_type", "target_value", ["devices"]
        )

        mock_find.assert_called_once_with(
            "tree", "level", "target_type", "target_value", ["devices"]
        )
        assert_report_item_list_equal(report_list, [])

    def test_error(self, mock_find):
        mock_find.return_value = ["element"]

        report_list = lib._validate_level_target_devices_does_not_exist(
            "tree", "level", "target_type", "target_value", ["devices"]
        )

        mock_find.assert_called_once_with(
            "tree", "level", "target_type", "target_value", ["devices"]
        )
        report = [
            (
                severity.ERROR,
                report_codes.CIB_FENCING_LEVEL_ALREADY_EXISTS,
                {
                    "devices": ["devices"],
                    "target_type": "target_type",
                    "target_value": "target_value",
                    "level": "level",
                },
                None,
            )
        ]
        assert_report_item_list_equal(report_list, report)


class AppendLevelElement(TestCase):
    def setUp(self):
        self.tree = etree.fromstring("<fencing-topology />")

    def test_node_name(self):
        lib._append_level_element(
            self.tree, 1, TARGET_TYPE_NODE, "node1", ["d1"]
        )
        assert_xml_equal(
            """
            <fencing-topology>
                <fencing-level
                    id="fl-node1-1"
                    devices="d1" index="1" target="node1"
                />
            </fencing-topology>
            """,
            etree_to_str(self.tree),
        )

    def test_node_pattern(self):
        lib._append_level_element(
            self.tree, "2", TARGET_TYPE_REGEXP, r"node-\d+", ["d1", "d2"]
        )
        assert_xml_equal(
            """
            <fencing-topology>
                <fencing-level
                    id="fl-node-d-2"
                    devices="d1,d2" index="2" target-pattern="node-\\d+"
                />
            </fencing-topology>
            """,
            etree_to_str(self.tree),
        )

    def test_node_attribute(self):
        lib._append_level_element(
            self.tree,
            3,
            TARGET_TYPE_ATTRIBUTE,
            ("name%@x", "val%@x"),
            ["d1"],
        )
        assert_xml_equal(
            """
            <fencing-topology>
                <fencing-level
                    id="fl-namex-3"
                    devices="d1"
                    index="3"
                    target-attribute="name%@x" target-value="val%@x"
                />
            </fencing-topology>
            """,
            etree_to_str(self.tree),
        )


class FindLevelElements(TestCase, CibMixin):
    def setUp(self):
        self.cib = self.get_cib()
        self.tree = self.cib.find("configuration/fencing-topology")

    def get_ids(self, elements):
        return [el.get("id") for el in elements]

    def test_no_filter(self):
        self.assertEqual(
            self.get_ids(lib._find_level_elements(self.tree)),
            [
                "fl1",
                "fl2",
                "fl3",
                "fl4",
                "fl5",
                "fl6",
                "fl7",
                "fl8",
                "fl9",
                "fl10",
            ],
        )

    def test_no_such_level(self):
        self.assertEqual(
            self.get_ids(
                lib._find_level_elements(
                    self.tree,
                    level=2,
                    target_type=TARGET_TYPE_NODE,
                    target_value="nodeB",
                    devices=["d5"],
                )
            ),
            [],
        )

    def test_level(self):
        self.assertEqual(
            self.get_ids(lib._find_level_elements(self.tree, level=1)),
            ["fl1", "fl3", "fl5"],
        )

    def test_target_node(self):
        self.assertEqual(
            self.get_ids(
                lib._find_level_elements(
                    self.tree,
                    target_type=TARGET_TYPE_NODE,
                    target_value="nodeB",
                )
            ),
            ["fl3", "fl4"],
        )

    def test_target_pattern(self):
        self.assertEqual(
            self.get_ids(
                lib._find_level_elements(
                    self.tree,
                    target_type=TARGET_TYPE_REGEXP,
                    target_value="node-R.*",
                )
            ),
            ["fl9"],
        )

    def test_target_attribute(self):
        self.assertEqual(
            self.get_ids(
                lib._find_level_elements(
                    self.tree,
                    target_type=TARGET_TYPE_ATTRIBUTE,
                    target_value=("fencing", "improved"),
                )
            ),
            ["fl7", "fl8"],
        )

    def test_devices(self):
        self.assertEqual(
            self.get_ids(lib._find_level_elements(self.tree, devices=["d3"])),
            ["fl2", "fl4"],
        )

        self.assertEqual(
            self.get_ids(
                lib._find_level_elements(self.tree, devices=["d1", "d2"])
            ),
            ["fl1"],
        )

    def test_combination(self):
        self.assertEqual(
            self.get_ids(
                lib._find_level_elements(
                    self.tree, 2, TARGET_TYPE_NODE, "nodeB", ["d3"]
                )
            ),
            ["fl4"],
        )



# pylint: disable=too-many-lines
from unittest import TestCase

from pcs.common.reports import codes as report_codes
from pcs.lib.commands import resource

from pcs_test.tier0.lib.commands.tag.tag_common import fixture_tags_xml
from pcs_test.tools import fixture
from pcs_test.tools.command_env import get_env_tools

fixture_primitive_cib_managed = """
    <resources>
        <primitive class="ocf" id="A" provider="heartbeat" type="Dummy">
        </primitive>
    </resources>
"""
fixture_primitive_cib_managed_with_meta = """
    <resources>
        <primitive class="ocf" id="A" provider="heartbeat" type="Dummy">
            <meta_attributes id="A-meta_attributes" />
        </primitive>
    </resources>
"""
fixture_primitive_cib_unmanaged = """
    <resources>
        <primitive class="ocf" id="A" provider="heartbeat" type="Dummy">
            <meta_attributes id="A-meta_attributes">
                <nvpair id="A-meta_attributes-is-managed"
                    name="is-managed" value="false" />
            </meta_attributes>
        </primitive>
    </resources>
"""

fixture_primitive_cib_managed_op_enabled = """
    <resources>
        <primitive class="ocf" id="A" provider="heartbeat" type="Stateful">
            <operations>
                <op id="A-start" name="start" />
                <op id="A-stop" name="stop" />
                <op id="A-monitor-m" name="monitor" role="Master" />
                <op id="A-monitor-s" name="monitor" role="Slave" />
            </operations>
        </primitive>
    </resources>
"""
fixture_primitive_cib_managed_with_meta_op_enabled = """
    <resources>
        <primitive class="ocf" id="A" provider="heartbeat" type="Stateful">
            <meta_attributes id="A-meta_attributes" />
            <operations>
                <op id="A-start" name="start" />
                <op id="A-stop" name="stop" />
                <op id="A-monitor-m" name="monitor" role="Master" />
                <op id="A-monitor-s" name="monitor" role="Slave" />
            </operations>
        </primitive>
    </resources>
"""
fixture_primitive_cib_managed_with_meta_op_disabled = """
    <resources>
        <primitive class="ocf" id="A" provider="heartbeat" type="Stateful">
            <meta_attributes id="A-meta_attributes" />
            <operations>
                <op id="A-start" name="start" />
                <op id="A-stop" name="stop" />
                <op id="A-monitor-m" name="monitor" role="Master"
                    enabled="false" />
                <op id="A-monitor-s" name="monitor" role="Slave"
                    enabled="false" />
            </operations>
        </primitive>
    </resources>
"""
fixture_primitive_cib_unmanaged_op_enabled = """
    <resources>
        <primitive class="ocf" id="A" provider="heartbeat" type="Stateful">
            <meta_attributes id="A-meta_attributes">
                <nvpair id="A-meta_attributes-is-managed"
                    name="is-managed" value="false" />
            </meta_attributes>
            <operations>
                <op id="A-start" name="start" />
                <op id="A-stop" name="stop" />
                <op id="A-monitor-m" name="monitor" role="Master" />
                <op id="A-monitor-s" name="monitor" role="Slave" />
            </operations>
        </primitive>
    </resources>
"""
fixture_primitive_cib_unmanaged_op_disabled = """
    <resources>
        <primitive class="ocf" id="A" provider="heartbeat" type="Stateful">
            <meta_attributes id="A-meta_attributes">
                <nvpair id="A-meta_attributes-is-managed"
                    name="is-managed" value="false" />
            </meta_attributes>
            <operations>
                <op id="A-start" name="start" />
                <op id="A-stop" name="stop" />
                <op id="A-monitor-m" name="monitor" role="Master"
                    enabled="false" />
                <op id="A-monitor-s" name="monitor" role="Slave"
                    enabled="false" />
            </operations>
        </primitive>
    </resources>
"""


def get_fixture_group_cib(
    group_unmanaged=False,
    group_meta=False,
    primitive1_unmanaged=False,
    primitive1_meta=False,
    primitive2_unmanaged=False,
    primitive2_meta=False,
):
    parts = [
        """
        <resources><group id="A">
        <primitive id="A1" class="ocf" provider="heartbeat" type="Dummy">
    """
    ]
    if primitive1_unmanaged:
        parts.append(
            """
            <meta_attributes id="A1-meta_attributes">
                <nvpair id="A1-meta_attributes-is-managed"
                    name="is-managed" value="false" />
            </meta_attributes>
        """
        )
    elif primitive1_meta:
        parts.append("""<meta_attributes id="A1-meta_attributes" />""")
    parts.append(
        """
        </primitive>
        <primitive id="A2" class="ocf" provider="heartbeat" type="Dummy">
    """
    )
    if primitive2_unmanaged:
        parts.append(
            """
            <meta_attributes id="A2-meta_attributes">
                <nvpair id="A2-meta_attributes-is-managed"
                    name="is-managed" value="false" />
            </meta_attributes>
        """
        )
    elif primitive2_meta:
        parts.append("""<meta_attributes id="A2-meta_attributes" />""")
    parts.append("""</primitive>""")
    if group_unmanaged:
        parts.append(
            """
            <meta_attributes id="A-meta_attributes">
                <nvpair id="A-meta_attributes-is-managed"
                    name="is-managed" value="false" />
            </meta_attributes>
        """
        )
    elif group_meta:
        parts.append("""<meta_attributes id="A-meta_attributes" />""")
    parts.append("""</group></resources>""")
    return "".join(parts)


fixture_group_cib_managed = get_fixture_group_cib()
fixture_group_cib_managed_with_both_meta = get_fixture_group_cib(
    group_meta=True, primitive1_meta=True
)
fixture_group_cib_managed_with_resources_meta = get_fixture_group_cib(
    primitive1_meta=True, primitive2_meta=True
)
fixture_group_cib_unmanaged_resource = get_fixture_group_cib(
    primitive1_unmanaged=True
)
fixture_group_cib_unmanaged_resource_with_other_resource_meta = (
    get_fixture_group_cib(primitive1_unmanaged=True, primitive2_meta=True)
)
fixture_group_cib_unmanaged_resource_and_group = get_fixture_group_cib(
    group_unmanaged=True, primitive1_unmanaged=True
)
fixture_group_cib_unmanaged_all_resources = get_fixture_group_cib(
    primitive1_unmanaged=True, primitive2_unmanaged=True
)


def get_fixture_clone_cib(
    clone_unmanaged=False,
    clone_meta=False,
    primitive_unmanaged=False,
    primitive_meta=False,
):
    parts = ["""<resources><clone id="A-clone">"""]
    if clone_unmanaged:
        parts.append(
            """
            <meta_attributes id="A-clone-meta_attributes">
                <nvpair id="A-clone-meta_attributes-is-managed"
                    name="is-managed" value="false" />
            </meta_attributes>
        """
        )
    elif clone_meta:
        parts.append("""<meta_attributes id="A-clone-meta_attributes" />""")
    parts.append(
        """<primitive id="A" class="ocf" provider="heartbeat" type="Dummy">"""
    )
    if primitive_unmanaged:
        parts.append(
            """
            <meta_attributes id="A-meta_attributes">
                <nvpair id="A-meta_attributes-is-managed"
                    name="is-managed" value="false" />
            </meta_attributes>
        """
        )
    elif primitive_meta:
        parts.append("""<meta_attributes id="A-meta_attributes" />""")
    parts.append("""</primitive></clone></resources>""")
    return "".join(parts)


fixture_clone_cib_managed = get_fixture_clone_cib()
fixture_clone_cib_managed_with_meta_primitive = get_fixture_clone_cib(
    primitive_meta=True
)
fixture_clone_cib_managed_with_meta_clone = get_fixture_clone_cib(
    clone_meta=True
)
fixture_clone_cib_managed_with_meta_both = get_fixture_clone_cib(
    clone_meta=True, primitive_meta=True
)
fixture_clone_cib_unmanaged_clone = get_fixture_clone_cib(clone_unmanaged=True)
fixture_clone_cib_unmanaged_primitive = get_fixture_clone_cib(
    primitive_unmanaged=True
)
fixture_clone_cib_unmanaged_both = get_fixture_clone_cib(
    clone_unmanaged=True, primitive_unmanaged=True
)

fixture_clone_cib_managed_op_enabled = """
    <resources>
        <clone id="A-clone">
            <primitive id="A" class="ocf" provider="heartbeat" type="Dummy">
                <operations>
                    <op id="A-start" name="start" />
                    <op id="A-stop" name="stop" />
                    <op id="A-monitor" name="monitor"/>
                </operations>
            </primitive>
        </clone>
    </resources>
"""
fixture_clone_cib_unmanaged_primitive_op_disabled = """
    <resources>
        <clone id="A-clone">
            <primitive id="A" class="ocf" provider="heartbeat" type="Dummy">
                <meta_attributes id="A-meta_attributes">
                    <nvpair id="A-meta_attributes-is-managed"
                        name="is-managed" value="false" />
                </meta_attributes>
                <operations>
                    <op id="A-start" name="start" />
                    <op id="A-stop" name="stop" />
                    <op id="A-monitor" name="monitor" enabled="false"/>
                </operations>
            </primitive>
        </clone>
    </resources>
"""

fixture_master_cib_managed = """
    <resources>
        <master id="A-master">
            <primitive id="A" class="ocf" provider="heartbeat" type="Dummy">
            </primitive>
        </master>
    </resources>
"""
fixture_master_cib_managed_with_master_meta = """
    <resources>
        <master id="A-master">
            <meta_attributes id="A-master-meta_attributes" />
            <primitive id="A" class="ocf" provider="heartbeat" type="Dummy">
            </primitive>
        </master>
    </resources>
"""
fixture_master_cib_managed_with_primitive_meta = """
    <resources>
        <master id="A-master">
            <primitive id="A" class="ocf" provider="heartbeat" type="Dummy">
                <meta_attributes id="A-meta_attributes" />
            </primitive>
        </master>
    </resources>
"""
fixture_master_cib_managed_with_both_meta = """
    <resources>
        <master id="A-master">
            <meta_attributes id="A-master-meta_attributes" />
            <primitive id="A" class="ocf" provider="heartbeat" type="Dummy">
                <meta_attributes id="A-meta_attributes" />
            </primitive>
        </master>
    </resources>
"""
fixture_master_cib_unmanaged_master = """
    <resources>
        <master id="A-master">
            <meta_attributes id="A-master-meta_attributes">
                <nvpair id="A-master-meta_attributes-is-managed"
                    name="is-managed" value="false" />
            </meta_attributes>
            <primitive id="A" class="ocf" provider="heartbeat" type="Dummy">
            </primitive>
        </master>
    </resources>
"""
fixture_master_cib_unmanaged_primitive = """
    <resources>
        <master id="A-master">
            <primitive id="A" class="ocf" provider="heartbeat" type="Dummy">
                <meta_attributes id="A-meta_attributes">
                    <nvpair id="A-meta_attributes-is-managed"
                        name="is-managed" value="false" />
                </meta_attributes>
            </primitive>
        </master>
    </resources>
"""
fixture_master_cib_unmanaged_both = """
    <resources>
        <master id="A-master">
            <meta_attributes id="A-master-meta_attributes">
                <nvpair id="A-master-meta_attributes-is-managed"
                    name="is-managed" value="false" />
            </meta_attributes>
            <primitive id="A" class="ocf" provider="heartbeat" type="Dummy">
                <meta_attributes id="A-meta_attributes">
                    <nvpair id="A-meta_attributes-is-managed"
                        name="is-managed" value="false" />
                </meta_attributes>
            </primitive>
        </master>
    </resources>
"""

fixture_master_cib_managed_op_enabled = """
    <resources>
        <master id="A-master">
            <primitive id="A" class="ocf" provider="heartbeat" type="Dummy">
                <operations>
                    <op id="A-start" name="start" />
                    <op id="A-stop" name="stop" />
                    <op id="A-monitor" name="monitor"/>
                </operations>
            </primitive>
        </master>
    </resources>
"""
fixture_master_cib_unmanaged_primitive_op_disabled = """
    <resources>
        <master id="A-master">
            <primitive id="A" class="ocf" provider="heartbeat" type="Dummy">
                <meta_attributes id="A-meta_attributes">
                    <nvpair id="A-meta_attributes-is-managed"
                        name="is-managed" value="false" />
                </meta_attributes>
                <operations>
                    <op id="A-start" name="start" />
                    <op id="A-stop" name="stop" />
                    <op id="A-monitor" name="monitor" enabled="false"/>
                </operations>
            </primitive>
        </master>
    </resources>
"""


def get_fixture_clone_group_cib(
    clone_unmanaged=False,
    clone_meta=False,
    group_unmanaged=False,
    group_meta=False,
    primitive1_unmanaged=False,
    primitive1_meta=False,
    primitive2_unmanaged=False,
    primitive2_meta=False,
    all_meta=False,
):
    # pylint: disable=too-many-arguments
    parts = ["""<resources><clone id="A-clone">"""]
    if clone_unmanaged:
        parts.append(
            """
            <meta_attributes id="A-clone-meta_attributes">
                <nvpair id="A-clone-meta_attributes-is-managed"
                    name="is-managed" value="false" />
            </meta_attributes>
        """
        )
    elif clone_meta or all_meta:
        parts.append("""<meta_attributes id="A-clone-meta_attributes" />""")
    parts.append("""<group id="A">""")
    if group_unmanaged:
        parts.append(
            """
            <meta_attributes id="A-meta_attributes">
                <nvpair id="A-meta_attributes-is-managed"
                    name="is-managed" value="false" />
            </meta_attributes>
        """
        )
    elif group_meta or all_meta:
        parts.append("""<meta_attributes id="A-meta_attributes" />""")
    parts.append(
        """<primitive id="A1" class="ocf" provider="heartbeat" type="Dummy">"""
    )
    if primitive1_unmanaged:
        parts.append(
            """
            <meta_attributes id="A1-meta_attributes">
                <nvpair id="A1-meta_attributes-is-managed"
                    name="is-managed" value="false" />
            </meta_attributes>
        """
        )
    elif primitive1_meta or all_meta:
        parts.append("""<meta_attributes id="A1-meta_attributes" />""")
    parts.append(
        """
        </primitive>
        <primitive id="A2" class="ocf" provider="heartbeat" type="Dummy">
    """
    )
    if primitive2_unmanaged:
        parts.append(
            """
            <meta_attributes id="A2-meta_attributes">
                <nvpair id="A2-meta_attributes-is-managed"
                    name="is-managed" value="false" />
            </meta_attributes>
        """
        )
    elif primitive2_meta or all_meta:
        parts.append("""<meta_attributes id="A2-meta_attributes" />""")
    parts.append("""</primitive></group></clone></resources>""")
    return "".join(parts)


fixture_clone_group_cib_managed = get_fixture_clone_group_cib()
fixture_clone_group_cib_managed_with_all_meta = get_fixture_clone_group_cib(
    all_meta=True
)
fixture_clone_group_cib_managed_with_all_primitive_meta = (
    get_fixture_clone_group_cib(primitive1_meta=True, primitive2_meta=True)
)
fixture_clone_group_cib_managed_with_clone_meta = get_fixture_clone_group_cib(
    clone_meta=True
)
fixture_clone_group_cib_managed_with_primitive_meta = (
    get_fixture_clone_group_cib(primitive1_meta=True)
)
fixture_clone_group_cib_unmanaged_primitive = get_fixture_clone_group_cib(
    primitive1_unmanaged=True
)
fixture_clone_group_cib_unmanaged_primitive_with_all_meta = (
    get_fixture_clone_group_cib(primitive1_unmanaged=True, all_meta=True)
)
fixture_clone_group_cib_unmanaged_all_primitives = get_fixture_clone_group_cib(
    primitive1_unmanaged=True, primitive2_unmanaged=True
)
fixture_clone_group_cib_unmanaged_clone = get_fixture_clone_group_cib(
    clone_unmanaged=True
)
fixture_clone_group_cib_unmanaged_everything = get_fixture_clone_group_cib(
    clone_unmanaged=True,
    group_unmanaged=True,
    primitive1_unmanaged=True,
    primitive2_unmanaged=True,
)

fixture_clone_group_cib_managed_op_enabled = """
    <resources>
        <clone id="A-clone">
            <group id="A">
                <primitive id="A1" class="ocf" provider="heartbeat"
                    type="Dummy"
                >
                    <operations>
                        <op id="A1-start" name="start" />
                        <op id="A1-stop" name="stop" />
                        <op id="A1-monitor" name="monitor" />
                    </operations>
                </primitive>
                <primitive id="A2" class="ocf" provider="heartbeat"
                    type="Dummy"
                >
                    <operations>
                        <op id="A2-start" name="start" />
                        <op id="A2-stop" name="stop" />
                        <op id="A2-monitor" name="monitor" />
                    </operations>
                </primitive>
            </group>
        </clone>
    </resources>
"""
fixture_clone_group_cib_unmanaged_primitive_op_disabled = """
    <resources>
        <clone id="A-clone">
            <group id="A">
                <primitive id="A1" class="ocf" provider="heartbeat"
                    type="Dummy"
                >
                    <meta_attributes id="A1-meta_attributes">
                        <nvpair id="A1-meta_attributes-is-managed"
                            name="is-managed" value="false" />
                    </meta_attributes>
                    <operations>
                        <op id="A1-start" name="start" />
                        <op id="A1-stop" name="stop" />
                        <op id="A1-monitor" name="monitor" enabled="false" />
                    </operations>
                </primitive>
                <primitive id="A2" class="ocf" provider="heartbeat"
                    type="Dummy"
                >
                    <operations>
                        <op id="A2-start" name="start" />
                        <op id="A2-stop" name="stop" />
                        <op id="A2-monitor" name="monitor" />
                    </operations>
                </primitive>
            </group>
        </clone>
    </resources>
"""
fixture_clone_group_cib_unmanaged_all_primitives_op_disabled = """
    <resources>
        <clone id="A-clone">
            <group id="A">
                <primitive id="A1" class="ocf" provider="heartbeat"
                    type="Dummy"
                >
                    <meta_attributes id="A1-meta_attributes">
                        <nvpair id="A1-meta_attributes-is-managed"
                            name="is-managed" value="false" />
                    </meta_attributes>
                    <operations>
                        <op id="A1-start" name="start" />
                        <op id="A1-stop" name="stop" />
                        <op id="A1-monitor" name="monitor" enabled="false" />
                    </operations>
                </primitive>
                <primitive id="A2" class="ocf" provider="heartbeat"
                    type="Dummy"
                >
                    <meta_attributes id="A2-meta_attributes">
                        <nvpair id="A2-meta_attributes-is-managed"
                            name="is-managed" value="false" />
                    </meta_attributes>
                    <operations>
                        <op id="A2-start" name="start" />
                        <op id="A2-stop" name="stop" />
                        <op id="A2-monitor" name="monitor" enabled="false" />
                    </operations>
                </primitive>
            </group>
        </clone>
    </resources>
"""


fixture_bundle_empty_cib_managed = """
    <resources>
        <bundle id="A-bundle">
            <docker image="pcs:test" />
        </bundle>
    </resources>
"""
fixture_bundle_empty_cib_managed_with_meta = """
    <resources>
        <bundle id="A-bundle">
            <meta_attributes id="A-bundle-meta_attributes" />
            <docker image="pcs:test" />
        </bundle>
    </resources>
"""
fixture_bundle_empty_cib_unmanaged_bundle = """
    <resources>
        <bundle id="A-bundle">
            <meta_attributes id="A-bundle-meta_attributes">
                <nvpair id="A-bundle-meta_attributes-is-managed"
                    name="is-managed" value="false" />
            </meta_attributes>
            <docker image="pcs:test" />
        </bundle>
    </resources>
"""


def get_fixture_bundle_cib(
    bundle_unmanaged=False,
    bundle_meta=False,
    primitive_unmanaged=False,
    primitive_meta=False,
):
    parts = ["""<resources><bundle id="A-bundle">"""]
    if bundle_unmanaged:
        parts.append(
            """
            <meta_attributes id="A-bundle-meta_attributes">
                <nvpair id="A-bundle-meta_attributes-is-managed"
                    name="is-managed" value="false" />
            </meta_attributes>
        """
        )
    elif bundle_meta:
        parts.append("""<meta_attributes id="A-bundle-meta_attributes" />""")
    parts.append(
        """
        <docker image="pcs:test" />
        <primitive id="A" class="ocf" provider="heartbeat" type="Dummy">
    """
    )
    if primitive_unmanaged:
        parts.append(
            """
            <meta_attributes id="A-meta_attributes">
                <nvpair id="A-meta_attributes-is-managed"
                    name="is-managed" value="false" />
            </meta_attributes>
        """
        )
    elif primitive_meta:
        parts.append("""<meta_attributes id="A-meta_attributes" />""")
    parts.append("""</primitive></bundle></resources>""")
    return "".join(parts)


fixture_bundle_cib_managed = get_fixture_bundle_cib()
fixture_bundle_cib_managed_with_meta_bundle = get_fixture_bundle_cib(
    bundle_meta=True
)
fixture_bundle_cib_managed_with_meta_both = get_fixture_bundle_cib(
    bundle_meta=True, primitive_meta=True
)
fixture_bundle_cib_managed_with_meta_primitive = get_fixture_bundle_cib(
    primitive_meta=True
)
fixture_bundle_cib_unmanaged_bundle = get_fixture_bundle_cib(
    bundle_unmanaged=True
)
fixture_bundle_cib_unmanaged_primitive = get_fixture_bundle_cib(
    primitive_unmanaged=True
)
fixture_bundle_cib_unmanaged_both = get_fixture_bundle_cib(
    bundle_unmanaged=True, primitive_unmanaged=True
)

fixture_bundle_cib_managed_op_enabled = """
    <resources>
        <bundle id="A-bundle">
            <docker image="pcs:test" />
            <primitive id="A" class="ocf" provider="heartbeat" type="Dummy">
                <operations>
                    <op id="A-start" name="start" />
                    <op id="A-stop" name="stop" />
                    <op id="A-monitor" name="monitor"/>
                </operations>
            </primitive>
        </bundle>
    </resources>
"""
fixture_bundle_cib_unmanaged_primitive_op_disabled = """
    <resources>
        <bundle id="A-bundle">
            <docker image="pcs:test" />
            <primitive id="A" class="ocf" provider="heartbeat" type="Dummy">
                <meta_attributes id="A-meta_attributes">
                    <nvpair id="A-meta_attributes-is-managed"
                        name="is-managed" value="false" />
                </meta_attributes>
                <operations>
                    <op id="A-start" name="start" />
                    <op id="A-stop" name="stop" />
                    <op id="A-monitor" name="monitor" enabled="false"/>
                </operations>
            </primitive>
        </bundle>
    </resources>
"""
fixture_bundle_cib_unmanaged_both_op_disabled = """
    <resources>
        <bundle id="A-bundle">
            <meta_attributes id="A-bundle-meta_attributes">
                <nvpair id="A-bundle-meta_attributes-is-managed"
                    name="is-managed" value="false" />
            </meta_attributes>
            <docker image="pcs:test" />
            <primitive id="A" class="ocf" provider="heartbeat" type="Dummy">
                <meta_attributes id="A-meta_attributes">
                    <nvpair id="A-meta_attributes-is-managed"
                        name="is-managed" value="false" />
                </meta_attributes>
                <operations>
                    <op id="A-start" name="start" />
                    <op id="A-stop" name="stop" />
                    <op id="A-monitor" name="monitor" enabled="false"/>
                </operations>
            </primitive>
        </bundle>
    </resources>
"""


def fixture_report_no_monitors(resource_id):
    return fixture.warn(
        report_codes.RESOURCE_MANAGED_NO_MONITOR_ENABLED,
        resource_id=resource_id,
    )


class UnmanageTag(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(test_case=self)

    def test_tag(self):
        self.config.runner.cib.load(
            resources=fixture_primitive_cib_managed,
            tags=fixture_tags_xml([("T", ("A"))]),
        )
        self.config.env.push_cib(resources=fixture_primitive_cib_unmanaged)
        resource.unmanage(self.env_assist.get_env(), ["T"])


class ManageTag(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(test_case=self)

    def test_tag(self):
        self.config.runner.cib.load(
            resources=fixture_primitive_cib_unmanaged,
            tags=fixture_tags_xml([("T", ("A"))]),
        )
        self.config.env.push_cib(
            resources=fixture_primitive_cib_managed_with_meta
        )
        resource.manage(self.env_assist.get_env(), ["T"])


class UnmanagePrimitive(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(test_case=self)

    def test_nonexistent_resource(self):
        self.config.runner.cib.load(resources=fixture_primitive_cib_managed)

        self.env_assist.assert_raise_library_error(
            lambda: resource.unmanage(self.env_assist.get_env(), ["B"]),
        )
        self.env_assist.assert_reports(
            [fixture.report_not_resource_or_tag("B")]
        )

    def test_primitive(self):
        self.config.runner.cib.load(resources=fixture_primitive_cib_managed)
        self.config.env.push_cib(resources=fixture_primitive_cib_unmanaged)
        resource.unmanage(self.env_assist.get_env(), ["A"])

    def test_primitive_unmanaged(self):
        self.config.runner.cib.load(resources=fixture_primitive_cib_unmanaged)
        self.config.env.push_cib(resources=fixture_primitive_cib_unmanaged)
        resource.unmanage(self.env_assist.get_env(), ["A"])


class ManagePrimitive(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(test_case=self)

    def test_nonexistent_resource(self):
        self.config.runner.cib.load(resources=fixture_primitive_cib_unmanaged)

        self.env_assist.assert_raise_library_error(
            lambda: resource.manage(self.env_assist.get_env(), ["B"]),
        )
        self.env_assist.assert_reports(
            [fixture.report_not_resource_or_tag("B")]
        )

    def test_primitive(self):
        self.config.runner.cib.load(resources=fixture_primitive_cib_unmanaged)
        self.config.env.push_cib(
            resources=fixture_primitive_cib_managed_with_meta
        )
        resource.manage(self.env_assist.get_env(), ["A"])

    def test_primitive_managed(self):
        self.config.runner.cib.load(resources=fixture_primitive_cib_managed)
        self.config.env.push_cib(resources=fixture_primitive_cib_managed)
        resource.manage(self.env_assist.get_env(), ["A"])


class UnmanageGroup(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(test_case=self)

    def test_primitive(self):
        self.config.runner.cib.load(resources=fixture_group_cib_managed)
        self.config.env.push_cib(resources=fixture_group_cib_unmanaged_resource)
        resource.unmanage(self.env_assist.get_env(), ["A1"])

    def test_group(self):
        self.config.runner.cib.load(resources=fixture_group_cib_managed)
        self.config.env.push_cib(
            resources=fixture_group_cib_unmanaged_all_resources
        )
        resource.unmanage(self.env_assist.get_env(), ["A"])


class ManageGroup(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(test_case=self)

    def test_primitive(self):
        self.config.runner.cib.load(
            resources=fixture_group_cib_unmanaged_all_resources
        )
        self.config.env.push_cib(
            resources=fixture_group_cib_unmanaged_resource_with_other_resource_meta
        )
        resource.manage(self.env_assist.get_env(), ["A2"])

    def test_primitive_unmanaged_group(self):
        self.config.runner.cib.load(
            resources=fixture_group_cib_unmanaged_resource_and_group
        )
        self.config.env.push_cib(
            resources=fixture_group_cib_managed_with_both_meta
        )
        resource.manage(self.env_assist.get_env(), ["A1"])

    def test_group(self):
        self.config.runner.cib.load(
            resources=fixture_group_cib_unmanaged_all_resources
        )
        self.config.env.push_cib(
            resources=fixture_group_cib_managed_with_resources_meta
        )
        resource.manage(self.env_assist.get_env(), ["A"])

    def test_group_unmanaged_group(self):
        self.config.runner.cib.load(
            resources=fixture_group_cib_unmanaged_resource_and_group
        )
        self.config.env.push_cib(
            resources=fixture_group_cib_managed_with_both_meta
        )
        resource.manage(self.env_assist.get_env(), ["A"])


class UnmanageClone(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(test_case=self)

    def test_primitive(self):
        self.config.runner.cib.load(resources=fixture_clone_cib_managed)
        self.config.env.push_cib(
            resources=fixture_clone_cib_unmanaged_primitive
        )
        resource.unmanage(self.env_assist.get_env(), ["A"])

    def test_clone(self):
        self.config.runner.cib.load(resources=fixture_clone_cib_managed)
        self.config.env.push_cib(
            resources=fixture_clone_cib_unmanaged_primitive
        )
        resource.unmanage(self.env_assist.get_env(), ["A-clone"])


class ManageClone(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(test_case=self)

    def test_primitive(self):
        self.config.runner.cib.load(resources=fixture_clone_cib_unmanaged_clone)
        self.config.env.push_cib(
            resources=fixture_clone_cib_managed_with_meta_clone
        )
        resource.manage(self.env_assist.get_env(), ["A"])

    def test_primitive_unmanaged_primitive(self):
        self.config.runner.cib.load(
            resources=fixture_clone_cib_unmanaged_primitive
        )
        self.config.env.push_cib(
            resources=fixture_clone_cib_managed_with_meta_primitive
        )
        resource.manage(self.env_assist.get_env(), ["A"])

    def test_primitive_unmanaged_both(self):
        self.config.runner.cib.load(resources=fixture_clone_cib_unmanaged_both)
        self.config.env.push_cib(
            resources=fixture_clone_cib_managed_with_meta_both
        )
        resource.manage(self.env_assist.get_env(), ["A"])

    def test_clone(self):
        self.config.runner.cib.load(resources=fixture_clone_cib_unmanaged_clone)
        self.config.env.push_cib(
            resources=fixture_clone_cib_managed_with_meta_clone
        )
        resource.manage(self.env_assist.get_env(), ["A-clone"])

    def test_clone_unmanaged_primitive(self):
        self.config.runner.cib.load(
            resources=fixture_clone_cib_unmanaged_primitive
        )
        self.config.env.push_cib(
            resources=fixture_clone_cib_managed_with_meta_primitive
        )
        resource.manage(self.env_assist.get_env(), ["A-clone"])

    def test_clone_unmanaged_both(self):
        self.config.runner.cib.load(resources=fixture_clone_cib_unmanaged_both)
        self.config.env.push_cib(
            resources=fixture_clone_cib_managed_with_meta_both
        )
        resource.manage(self.env_assist.get_env(), ["A-clone"])


class UnmanageMaster(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(test_case=self)

    def test_primitive(self):
        self.config.runner.cib.load(resources=fixture_master_cib_managed)
        self.config.env.push_cib(
            resources=fixture_master_cib_unmanaged_primitive
        )
        resource.unmanage(self.env_assist.get_env(), ["A"])

    def test_master(self):
        self.config.runner.cib.load(resources=fixture_master_cib_managed)
        self.config.env.push_cib(
            resources=fixture_master_cib_unmanaged_primitive
        )
        resource.unmanage(self.env_assist.get_env(), ["A-master"])


class ManageMaster(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(test_case=self)

    def test_primitive(self):
        self.config.runner.cib.load(
            resources=fixture_master_cib_unmanaged_primitive
        )
        self.config.env.push_cib(
            resources=fixture_master_cib_managed_with_primitive_meta
        )
        resource.manage(self.env_assist.get_env(), ["A"])

    def test_primitive_unmanaged_master(self):
        self.config.runner.cib.load(
            resources=fixture_master_cib_unmanaged_master
        )
        self.config.env.push_cib(
            resources=fixture_master_cib_managed_with_master_meta
        )
        resource.manage(self.env_assist.get_env(), ["A"])

    def test_primitive_unmanaged_both(self):
        self.config.runner.cib.load(resources=fixture_master_cib_unmanaged_both)
        self.config.env.push_cib(
            resources=fixture_master_cib_managed_with_both_meta
        )
        resource.manage(self.env_assist.get_env(), ["A"])

    def test_master(self):
        self.config.runner.cib.load(
            resources=fixture_master_cib_unmanaged_master
        )
        self.config.env.push_cib(
            resources=fixture_master_cib_managed_with_master_meta
        )
        resource.manage(self.env_assist.get_env(), ["A-master"])

    def test_master_unmanaged_primitive(self):
        self.config.runner.cib.load(
            resources=fixture_master_cib_unmanaged_primitive
        )
        self.config.env.push_cib(
            resources=fixture_master_cib_managed_with_primitive_meta
        )
        resource.manage(self.env_assist.get_env(), ["A-master"])

    def test_master_unmanaged_both(self):
        self.config.runner.cib.load(resources=fixture_master_cib_unmanaged_both)
        self.config.env.push_cib(
            resources=fixture_master_cib_managed_with_both_meta
        )
        resource.manage(self.env_assist.get_env(), ["A-master"])


class UnmanageClonedGroup(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(test_case=self)

    def test_primitive(self):
        self.config.runner.cib.load(resources=fixture_clone_group_cib_managed)
        self.config.env.push_cib(
            resources=fixture_clone_group_cib_unmanaged_primitive
        )
        resource.unmanage(self.env_assist.get_env(), ["A1"])

    def test_group(self):
        self.config.runner.cib.load(resources=fixture_clone_group_cib_managed)
        self.config.env.push_cib(
            resources=fixture_clone_group_cib_unmanaged_all_primitives
        )
        resource.unmanage(self.env_assist.get_env(), ["A"])

    def test_clone(self):
        self.config.runner.cib.load(resources=fixture_clone_group_cib_managed)
        self.config.env.push_cib(
            resources=fixture_clone_group_cib_unmanaged_all_primitives
        )
        resource.unmanage(self.env_assist.get_env(), ["A-clone"])


class ManageClonedGroup(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(test_case=self)

    def test_primitive(self):
        self.config.runner.cib.load(
            resources=fixture_clone_group_cib_unmanaged_primitive
        )
        self.config.env.push_cib(
            resources=fixture_clone_group_cib_managed_with_primitive_meta
        )
        resource.manage(self.env_assist.get_env(), ["A1"])

    def test_primitive_unmanaged_all(self):
        self.config.runner.cib.load(
            resources=fixture_clone_group_cib_unmanaged_everything
        )
        self.config.env.push_cib(
            resources=fixture_clone_group_cib_unmanaged_primitive_with_all_meta
        )
        resource.manage(self.env_assist.get_env(), ["A2"])

    def test_group(self):
        self.config.runner.cib.load(
            resources=fixture_clone_group_cib_unmanaged_all_primitives
        )
        self.config.env.push_cib(
            resources=fixture_clone_group_cib_managed_with_all_primitive_meta
        )
        resource.manage(self.env_assist.get_env(), ["A"])

    def test_group_unmanaged_all(self):
        self.config.runner.cib.load(
            resources=fixture_clone_group_cib_unmanaged_everything
        )
        self.config.env.push_cib(
            resources=fixture_clone_group_cib_managed_with_all_meta
        )
        resource.manage(self.env_assist.get_env(), ["A"])

    def test_clone(self):
        self.config.runner.cib.load(
            resources=fixture_clone_group_cib_unmanaged_clone
        )
        self.config.env.push_cib(
            resources=fixture_clone_group_cib_managed_with_clone_meta
        )
        resource.manage(self.env_assist.get_env(), ["A-clone"])

    def test_clone_unmanaged_all(self):
        self.config.runner.cib.load(
            resources=fixture_clone_group_cib_unmanaged_everything
        )
        self.config.env.push_cib(
            resources=fixture_clone_group_cib_managed_with_all_meta
        )
        resource.manage(self.env_assist.get_env(), ["A-clone"])


class UnmanageBundle(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(test_case=self)

    def test_primitive(self):
        self.config.runner.cib.load(resources=fixture_bundle_cib_managed)
        self.config.env.push_cib(
            resources=fixture_bundle_cib_unmanaged_primitive
        )
        resource.unmanage(self.env_assist.get_env(), ["A"])

    def test_bundle(self):
        self.config.runner.cib.load(resources=fixture_bundle_cib_managed)
        self.config.env.push_cib(resources=fixture_bundle_cib_unmanaged_both)
        resource.unmanage(self.env_assist.get_env(), ["A-bundle"])

    def test_bundle_empty(self):
        self.config.runner.cib.load(resources=fixture_bundle_empty_cib_managed)
        self.config.env.push_cib(
            resources=fixture_bundle_empty_cib_unmanaged_bundle
        )
        resource.unmanage(self.env_assist.get_env(), ["A-bundle"])


class ManageBundle(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(test_case=self)

    def test_primitive(self):
        self.config.runner.cib.load(
            resources=fixture_bundle_cib_unmanaged_primitive
        )
        self.config.env.push_cib(
            resources=fixture_bundle_cib_managed_with_meta_primitive
        )
        resource.manage(self.env_assist.get_env(), ["A"])

    def test_primitive_unmanaged_bundle(self):
        self.config.runner.cib.load(
            resources=fixture_bundle_cib_unmanaged_bundle
        )
        self.config.env.push_cib(
            resources=fixture_bundle_cib_managed_with_meta_bundle
        )
        resource.manage(self.env_assist.get_env(), ["A"])

    def test_primitive_unmanaged_both(self):
        self.config.runner.cib.load(resources=fixture_bundle_cib_unmanaged_both)
        self.config.env.push_cib(
            resources=fixture_bundle_cib_managed_with_meta_both
        )
        resource.manage(self.env_assist.get_env(), ["A"])

    def test_bundle(self):
        self.config.runner.cib.load(
            resources=fixture_bundle_cib_unmanaged_bundle
        )
        self.config.env.push_cib(
            resources=fixture_bundle_cib_managed_with_meta_bundle
        )
        resource.manage(self.env_assist.get_env(), ["A-bundle"])

    def test_bundle_unmanaged_primitive(self):
        self.config.runner.cib.load(
            resources=fixture_bundle_cib_unmanaged_primitive
        )
        self.config.env.push_cib(
            resources=fixture_bundle_cib_managed_with_meta_primitive
        )
        resource.manage(self.env_assist.get_env(), ["A-bundle"])

    def test_bundle_unmanaged_both(self):
        self.config.runner.cib.load(resources=fixture_bundle_cib_unmanaged_both)
        self.config.env.push_cib(
            resources=fixture_bundle_cib_managed_with_meta_both
        )
        resource.manage(self.env_assist.get_env(), ["A-bundle"])

    def test_bundle_empty(self):
        self.config.runner.cib.load(
            resources=fixture_bundle_empty_cib_unmanaged_bundle
        )
        self.config.env.push_cib(
            resources=fixture_bundle_empty_cib_managed_with_meta
        )
        resource.manage(self.env_assist.get_env(), ["A-bundle"])


class MoreResources(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(test_case=self)

    fixture_cib_managed = """
        <resources>
            <primitive class="ocf" id="A" provider="heartbeat" type="Dummy">
            </primitive>
            <primitive class="ocf" id="B" provider="heartbeat" type="Dummy">
            </primitive>
            <primitive class="ocf" id="C" provider="heartbeat" type="Dummy">
            </primitive>
        </resources>
    """
    fixture_cib_unmanaged = """
        <resources>
            <primitive class="ocf" id="A" provider="heartbeat" type="Dummy">
                <meta_attributes id="A-meta_attributes">
                    <nvpair id="A-meta_attributes-is-managed"
                        name="is-managed" value="false" />
                </meta_attributes>
            </primitive>
            <primitive class="ocf" id="B" provider="heartbeat" type="Dummy">
                <meta_attributes id="B-meta_attributes">
                    <nvpair id="B-meta_attributes-is-managed"
                        name="is-managed" value="false" />
                </meta_attributes>
            </primitive>
            <primitive class="ocf" id="C" provider="heartbeat" type="Dummy">
                <meta_attributes id="C-meta_attributes">
                    <nvpair id="C-meta_attributes-is-managed"
                        name="is-managed" value="false" />
                </meta_attributes>
            </primitive>
        </resources>
    """

    def test_success_unmanage(self):
        fixture_cib_unmanaged = """
            <resources>
                <primitive class="ocf" id="A" provider="heartbeat" type="Dummy">
                    <meta_attributes id="A-meta_attributes">
                        <nvpair id="A-meta_attributes-is-managed"
                            name="is-managed" value="false" />
                    </meta_attributes>
                </primitive>
                <primitive class="ocf" id="B" provider="heartbeat" type="Dummy">
                </primitive>
                <primitive class="ocf" id="C" provider="heartbeat" type="Dummy">
                    <meta_attributes id="C-meta_attributes">
                        <nvpair id="C-meta_attributes-is-managed"
                            name="is-managed" value="false" />
                    </meta_attributes>
                </primitive>
            </resources>
        """
        self.config.runner.cib.load(resources=self.fixture_cib_managed)
        self.config.env.push_cib(resources=fixture_cib_unmanaged)
        resource.unmanage(self.env_assist.get_env(), ["A", "C"])

    def test_success_manage(self):
        fixture_cib_managed = """
            <resources>
                <primitive class="ocf" id="A" provider="heartbeat" type="Dummy">
                    <meta_attributes id="A-meta_attributes" />
                </primitive>
                <primitive class="ocf" id="B" provider="heartbeat" type="Dummy">
                    <meta_attributes id="B-meta_attributes">
                        <nvpair id="B-meta_attributes-is-managed"
                            name="is-managed" value="false" />
                    </meta_attributes>
                </primitive>
                <primitive class="ocf" id="C" provider="heartbeat" type="Dummy">
                    <meta_attributes id="C-meta_attributes" />
                </primitive>
            </resources>
        """
        self.config.runner.cib.load(resources=self.fixture_cib_unmanaged)
        self.config.env.push_cib(resources=fixture_cib_managed)
        resource.manage(self.env_assist.get_env(), ["A", "C"])

    def test_bad_resource_unmanage(self):
        self.config.runner.cib.load(resources=self.fixture_cib_managed)

        self.env_assist.assert_raise_library_error(
            lambda: resource.unmanage(
                self.env_assist.get_env(), ["B", "X", "Y", "A"]
            ),
        )
        self.env_assist.assert_reports(
            [
                fixture.report_not_resource_or_tag("X"),
                fixture.report_not_resource_or_tag("Y"),
            ],
        )

    def test_bad_resource_enable(self):
        self.config.runner.cib.load(resources=self.fixture_cib_unmanaged)

        self.env_assist.assert_raise_library_error(
            lambda: resource.manage(
                self.env_assist.get_env(), ["B", "X", "Y", "A"]
            ),
        )
        self.env_assist.assert_reports(
            [
                fixture.report_not_resource_or_tag("X"),
                fixture.report_not_resource_or_tag("Y"),
            ],
        )


class WithMonitor(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(test_case=self)

    def test_unmanage_noop(self):
        self.config.runner.cib.load(resources=fixture_primitive_cib_managed)
        self.config.env.push_cib(resources=fixture_primitive_cib_unmanaged)
        resource.unmanage(self.env_assist.get_env(), ["A"], True)

    def test_manage_noop(self):
        self.config.runner.cib.load(resources=fixture_primitive_cib_unmanaged)
        self.config.env.push_cib(
            resources=fixture_primitive_cib_managed_with_meta
        )
        resource.manage(self.env_assist.get_env(), ["A"], True)

    def test_unmanage(self):
        self.config.runner.cib.load(
            resources=fixture_primitive_cib_managed_op_enabled
        )
        self.config.env.push_cib(
            resources=fixture_primitive_cib_unmanaged_op_disabled
        )
        resource.unmanage(self.env_assist.get_env(), ["A"], True)

    def test_manage(self):
        self.config.runner.cib.load(
            resources=fixture_primitive_cib_unmanaged_op_disabled
        )
        self.config.env.push_cib(
            resources=fixture_primitive_cib_managed_with_meta_op_enabled
        )
        resource.manage(self.env_assist.get_env(), ["A"], True)

    def test_unmanage_enabled_monitors(self):
        self.config.runner.cib.load(
            resources=fixture_primitive_cib_managed_op_enabled
        )
        self.config.env.push_cib(
            resources=fixture_primitive_cib_unmanaged_op_enabled
        )
        resource.unmanage(self.env_assist.get_env(), ["A"], False)

    def test_manage_disabled_monitors(self):
        self.config.runner.cib.load(
            resources=fixture_primitive_cib_unmanaged_op_disabled
        )
        self.config.env.push_cib(
            resources=fixture_primitive_cib_managed_with_meta_op_disabled
        )
        resource.manage(self.env_assist.get_env(), ["A"], False)
        self.env_assist.assert_reports([fixture_report_no_monitors("A")])

    def test_unmanage_clone(self):
        self.config.runner.cib.load(
            resources=fixture_clone_cib_managed_op_enabled
        )
        self.config.env.push_cib(
            resources=fixture_clone_cib_unmanaged_primitive_op_disabled
        )
        resource.unmanage(self.env_assist.get_env(), ["A-clone"], True)

    def test_unmanage_in_clone(self):
        self.config.runner.cib.load(
            resources=fixture_clone_cib_managed_op_enabled
        )
        self.config.env.push_cib(
            resources=fixture_clone_cib_unmanaged_primitive_op_disabled
        )
        resource.unmanage(self.env_assist.get_env(), ["A"], True)

    def test_unmanage_master(self):
        self.config.runner.cib.load(
            resources=fixture_master_cib_managed_op_enabled
        )
        self.config.env.push_cib(
            resources=fixture_master_cib_unmanaged_primitive_op_disabled
        )
        resource.unmanage(self.env_assist.get_env(), ["A-master"], True)

    def test_unmanage_in_master(self):
        self.config.runner.cib.load(
            resources=fixture_master_cib_managed_op_enabled
        )
        self.config.env.push_cib(
            resources=fixture_master_cib_unmanaged_primitive_op_disabled
        )
        resource.unmanage(self.env_assist.get_env(), ["A"], True)

    def test_unmanage_clone_with_group(self):
        self.config.runner.cib.load(
            resources=fixture_clone_group_cib_managed_op_enabled
        )
        self.config.env.push_cib(
            resources=fixture_clone_group_cib_unmanaged_all_primitives_op_disabled
        )
        resource.unmanage(self.env_assist.get_env(), ["A-clone"], True)

    def test_unmanage_group_in_clone(self):
        self.config.runner.cib.load(
            resources=fixture_clone_group_cib_managed_op_enabled
        )
        self.config.env.push_cib(
            resources=fixture_clone_group_cib_unmanaged_all_primitives_op_disabled
        )
        resource.unmanage(self.env_assist.get_env(), ["A"], True)

    def test_unmanage_in_cloned_group(self):
        self.config.runner.cib.load(
            resources=fixture_clone_group_cib_managed_op_enabled
        )
        self.config.env.push_cib(
            resources=fixture_clone_group_cib_unmanaged_primitive_op_disabled
        )
        resource.unmanage(self.env_assist.get_env(), ["A1"], True)

    def test_unmanage_bundle(self):
        self.config.runner.cib.load(
            resources=fixture_bundle_cib_managed_op_enabled
        )
        self.config.env.push_cib(
            resources=fixture_bundle_cib_unmanaged_both_op_disabled
        )
        resource.unmanage(self.env_assist.get_env(), ["A-bundle"], True)

    def test_unmanage_in_bundle(self):
        self.config.runner.cib.load(
            resources=fixture_bundle_cib_managed_op_enabled
        )
        self.config.env.push_cib(
            resources=fixture_bundle_cib_unmanaged_primitive_op_disabled
        )
        resource.unmanage(self.env_assist.get_env(), ["A"], True)

    def test_unmanage_bundle_empty(self):
        self.config.runner.cib.load(resources=fixture_bundle_empty_cib_managed)
        self.config.env.push_cib(
            resources=fixture_bundle_empty_cib_unmanaged_bundle
        )
        resource.unmanage(self.env_assist.get_env(), ["A-bundle"], True)

    def test_manage_one_resource_in_group(self):
        self.config.runner.cib.load(
            resources=fixture_clone_group_cib_unmanaged_all_primitives_op_disabled
        )
        self.config.env.push_cib(
            resources="""
                <resources>
                    <clone id="A-clone">
                        <group id="A">
                            <primitive id="A1" class="ocf" provider="heartbeat"
                                type="Dummy"
                            >
                                <meta_attributes id="A1-meta_attributes" />
                                <operations>
                                    <op id="A1-start" name="start" />
                                    <op id="A1-stop" name="stop" />
                                    <op id="A1-monitor" name="monitor" />
                                </operations>
                            </primitive>
                            <primitive id="A2" class="ocf" provider="heartbeat"
                                type="Dummy"
                            >
                                <meta_attributes id="A2-meta_attributes">
                                    <nvpair id="A2-meta_attributes-is-managed"
                                        name="is-managed" value="false" />
                                </meta_attributes>
                                <operations>
                                    <op id="A2-start" name="start" />
                                    <op id="A2-stop" name="stop" />
                                    <op id="A2-monitor" name="monitor"
                                        enabled="false"
                                    />
                                </operations>
                            </primitive>
                        </group>
                    </clone>
                </resources>
            """
        )
        resource.manage(self.env_assist.get_env(), ["A1"], True)



from pcs import settings
from pcs.common.types import StringCollection

DEFAULT_INSTANCE_NAME = "booth"
AUTHFILE_FIX_OPTION = "enable-authfile"

GLOBAL_KEYS: StringCollection = (
    "transport",
    "port",
    "name",
    "authfile",
    "maxtimeskew",
    "site",
    "arbitrator",
    "site-user",
    "site-group",
    "arbitrator-user",
    "arbitrator-group",
    "debug",
    "ticket",
) + (
    (AUTHFILE_FIX_OPTION,)
    if settings.booth_enable_authfile_set_enabled
    or settings.booth_enable_authfile_unset_enabled
    else tuple()
)

TICKET_KEYS = (
    "acquire-after",
    "attr-prereq",
    "before-acquire-handler",
    "expire",
    "mode",
    "renewal-freq",
    "retries",
    "timeout",
    "weights",
)



class Env:
    # pylint: disable=too-many-instance-attributes, too-few-public-methods
    def __init__(self):
        self.cib_data = None
        self.user = None
        self.groups = None
        self.corosync_conf_data = None
        self.booth = None
        self.pacemaker = None
        self.known_hosts_getter = None
        self.debug = False
        self.request_timeout = None
        self.report_processor = None



import logging
from functools import partial
from unittest import (
    TestCase,
    mock,
)

from pcs.common import file_type_codes
from pcs.common.reports import ReportItemSeverity as severity
from pcs.common.reports import codes as report_codes
from pcs.lib.env import LibraryEnvironment

from pcs_test.tools.assertions import assert_raise_library_error
from pcs_test.tools.custom_mock import MockLibraryReportProcessor
from pcs_test.tools.misc import create_patcher

patch_env = create_patcher("pcs.lib.env")
patch_env_object = partial(mock.patch.object, LibraryEnvironment)


class LibraryEnvironmentTest(TestCase):
    def setUp(self):
        self.mock_logger = mock.MagicMock(logging.Logger)
        self.mock_reporter = MockLibraryReportProcessor()

    def test_logger(self):
        env = LibraryEnvironment(self.mock_logger, self.mock_reporter)
        self.assertEqual(self.mock_logger, env.logger)

    def test_report_processor(self):
        env = LibraryEnvironment(self.mock_logger, self.mock_reporter)
        self.assertEqual(self.mock_reporter, env.report_processor)

    def test_user_set(self):
        user = "testuser"
        env = LibraryEnvironment(
            self.mock_logger, self.mock_reporter, user_login=user
        )
        self.assertEqual(user, env.user_login)

    def test_user_not_set(self):
        env = LibraryEnvironment(self.mock_logger, self.mock_reporter)
        self.assertEqual(None, env.user_login)

    def test_usergroups_set(self):
        groups = ["some", "group"]
        env = LibraryEnvironment(
            self.mock_logger, self.mock_reporter, user_groups=groups
        )
        self.assertEqual(groups, env.user_groups)

    def test_usergroups_not_set(self):
        env = LibraryEnvironment(self.mock_logger, self.mock_reporter)
        self.assertEqual([], env.user_groups)


class GhostFileCodes(TestCase):
    def setUp(self):
        self.mock_logger = mock.MagicMock(logging.Logger)
        self.mock_reporter = MockLibraryReportProcessor()

    def _fixture_get_env(self, cib_data=None, corosync_conf_data=None):
        return LibraryEnvironment(
            self.mock_logger,
            self.mock_reporter,
            cib_data=cib_data,
            corosync_conf_data=corosync_conf_data,
        )

    def test_nothing(self):
        self.assertEqual(self._fixture_get_env().ghost_file_codes, [])

    def test_corosync(self):
        self.assertEqual(
            self._fixture_get_env(corosync_conf_data="x").ghost_file_codes,
            [file_type_codes.COROSYNC_CONF],
        )

    def test_cib(self):
        self.assertEqual(
            self._fixture_get_env(cib_data="x").ghost_file_codes,
            [file_type_codes.CIB],
        )

    def test_all(self):
        self.assertEqual(
            self._fixture_get_env(
                cib_data="x",
                corosync_conf_data="x",
            ).ghost_file_codes,
            sorted([file_type_codes.COROSYNC_CONF, file_type_codes.CIB]),
        )


@patch_env("CommandRunner")
class CmdRunner(TestCase):
    def setUp(self):
        self.mock_logger = mock.MagicMock(logging.Logger)
        self.mock_reporter = MockLibraryReportProcessor()

    def test_no_options(self, mock_runner):
        expected_runner = mock.MagicMock()
        mock_runner.return_value = expected_runner
        env = LibraryEnvironment(self.mock_logger, self.mock_reporter)
        runner = env.cmd_runner()
        self.assertEqual(expected_runner, runner)
        mock_runner.assert_called_once_with(
            self.mock_logger,
            self.mock_reporter,
            {
                "LC_ALL": "C",
            },
        )

    def test_user(self, mock_runner):
        expected_runner = mock.MagicMock()
        mock_runner.return_value = expected_runner
        user = "testuser"
        env = LibraryEnvironment(
            self.mock_logger, self.mock_reporter, user_login=user
        )
        runner = env.cmd_runner()
        self.assertEqual(expected_runner, runner)
        mock_runner.assert_called_once_with(
            self.mock_logger,
            self.mock_reporter,
            {
                "CIB_user": user,
                "LC_ALL": "C",
            },
        )

    @patch_env("create_tmp_cib")
    def test_dump_cib_file(self, mock_tmpfile, mock_runner):
        tmp_file_name = "a file"
        expected_runner = mock.MagicMock()
        mock_runner.return_value = expected_runner
        mock_instance = mock.MagicMock()
        mock_instance.name = tmp_file_name
        mock_tmpfile.return_value = mock_instance
        env = LibraryEnvironment(
            self.mock_logger, self.mock_reporter, cib_data="<cib />"
        )
        runner = env.cmd_runner()
        self.assertEqual(expected_runner, runner)
        mock_runner.assert_called_once_with(
            self.mock_logger,
            self.mock_reporter,
            {
                "LC_ALL": "C",
                "CIB_file": tmp_file_name,
            },
        )
        mock_tmpfile.assert_called_once_with(self.mock_reporter, "<cib />")


@patch_env_object("cmd_runner", lambda self: "runner")
class EnsureValidWait(TestCase):
    def setUp(self):
        self.create_env = partial(
            LibraryEnvironment,
            mock.MagicMock(logging.Logger),
            MockLibraryReportProcessor(),
        )

    @property
    def env_live(self):
        return self.create_env()

    @property
    def env_fake(self):
        return self.create_env(cib_data="<cib/>")

    def test_not_raises_if_waiting_false_no_matter_if_env_is_live(self):
        self.env_live.ensure_wait_satisfiable(False)
        self.env_fake.ensure_wait_satisfiable(False)

    def test_raises_when_is_not_live(self):
        env = self.env_fake
        assert_raise_library_error(
            lambda: env.ensure_wait_satisfiable(10),
            (
                severity.ERROR,
                report_codes.WAIT_FOR_IDLE_NOT_LIVE_CLUSTER,
                {},
                None,
            ),
        )

    @patch_env("get_valid_timeout_seconds")
    def test_do_checks(self, get_valid_timeout):
        timeout = 10
        env = self.env_live
        get_valid_timeout.return_value = timeout
        env.ensure_wait_satisfiable(timeout)
        get_valid_timeout.assert_called_once_with(timeout)



# pylint: disable=wildcard-import, unused-wildcard-import
# Wildcard import of deprecated report codes will prevent creation of a new
# report with the code of deprecated report
from .deprecated_codes import *
from .types import ForceCode as F
from .types import MessageCode as M

# force categories
FORCE = F("FORCE")
SKIP_OFFLINE_NODES = F("SKIP_OFFLINE_NODES")


# messages


ADD_REMOVE_ITEMS_NOT_SPECIFIED = M("ADD_REMOVE_ITEMS_NOT_SPECIFIED")
ADD_REMOVE_ITEMS_DUPLICATION = M("ADD_REMOVE_ITEMS_DUPLICATION")
ADD_REMOVE_CANNOT_ADD_ITEMS_ALREADY_IN_THE_CONTAINER = M(
    "ADD_REMOVE_CANNOT_ADD_ITEMS_ALREADY_IN_THE_CONTAINER"
)
ADD_REMOVE_CANNOT_REMOVE_ITEMS_NOT_IN_THE_CONTAINER = M(
    "ADD_REMOVE_CANNOT_REMOVE_ITEMS_NOT_IN_THE_CONTAINER"
)
ADD_REMOVE_CANNOT_ADD_AND_REMOVE_ITEMS_AT_THE_SAME_TIME = M(
    "ADD_REMOVE_CANNOT_ADD_AND_REMOVE_ITEMS_AT_THE_SAME_TIME"
)
ADD_REMOVE_CANNOT_REMOVE_ALL_ITEMS_FROM_THE_CONTAINER = M(
    "ADD_REMOVE_CANNOT_REMOVE_ALL_ITEMS_FROM_THE_CONTAINER"
)
ADD_REMOVE_ADJACENT_ITEM_NOT_IN_THE_CONTAINER = M(
    "ADD_REMOVE_ADJACENT_ITEM_NOT_IN_THE_CONTAINER"
)
ADD_REMOVE_CANNOT_PUT_ITEM_NEXT_TO_ITSELF = M(
    "ADD_REMOVE_CANNOT_PUT_ITEM_NEXT_TO_ITSELF"
)
ADD_REMOVE_CANNOT_SPECIFY_ADJACENT_ITEM_WITHOUT_ITEMS_TO_ADD = M(
    "ADD_REMOVE_CANNOT_SPECIFY_ADJACENT_ITEM_WITHOUT_ITEMS_TO_ADD"
)
AGENT_GENERIC_ERROR = M("AGENT_GENERIC_ERROR")
AGENT_IMPLEMENTS_UNSUPPORTED_OCF_VERSION = M(
    "AGENT_IMPLEMENTS_UNSUPPORTED_OCF_VERSION"
)
AGENT_NAME_GUESS_FOUND_MORE_THAN_ONE = M("AGENT_NAME_GUESS_FOUND_MORE_THAN_ONE")
AGENT_NAME_GUESS_FOUND_NONE = M("AGENT_NAME_GUESS_FOUND_NONE")
AGENT_NAME_GUESSED = M("AGENT_NAME_GUESSED")
AGENT_SELF_VALIDATION_INVALID_DATA = M("AGENT_SELF_VALIDATION_INVALID_DATA")
AGENT_SELF_VALIDATION_SKIPPED_UPDATED_RESOURCE_MISCONFIGURED = M(
    "AGENT_SELF_VALIDATION_SKIPPED_UPDATED_RESOURCE_MISCONFIGURED"
)
AGENT_SELF_VALIDATION_RESULT = M("AGENT_SELF_VALIDATION_RESULT")
BAD_CLUSTER_STATE_FORMAT = M("BAD_CLUSTER_STATE_FORMAT")
BOOTH_ADDRESS_DUPLICATION = M("BOOTH_ADDRESS_DUPLICATION")
BOOTH_ALREADY_IN_CIB = M("BOOTH_ALREADY_IN_CIB")
BOOTH_AUTHFILE_NOT_USED = M("BOOTH_AUTHFILE_NOT_USED")
BOOTH_UNSUPPORTED_OPTION_ENABLE_AUTHFILE = M(
    "BOOTH_UNSUPPORTED_OPTION_ENABLE_AUTHFILE"
)
BOOTH_CANNOT_DETERMINE_LOCAL_SITE_IP = M("BOOTH_CANNOT_DETERMINE_LOCAL_SITE_IP")
BOOTH_CONFIG_ACCEPTED_BY_NODE = M("BOOTH_CONFIG_ACCEPTED_BY_NODE")
BOOTH_CONFIG_DISTRIBUTION_NODE_ERROR = M("BOOTH_CONFIG_DISTRIBUTION_NODE_ERROR")
BOOTH_CONFIG_DISTRIBUTION_STARTED = M("BOOTH_CONFIG_DISTRIBUTION_STARTED")
BOOTH_CONFIG_IS_USED = M("BOOTH_CONFIG_IS_USED")
BOOTH_CONFIG_UNEXPECTED_LINES = M("BOOTH_CONFIG_UNEXPECTED_LINES")
BOOTH_DAEMON_STATUS_ERROR = M("BOOTH_DAEMON_STATUS_ERROR")
BOOTH_EVEN_PEERS_NUM = M("BOOTH_EVEN_PEERS_NUM")
BOOTH_FETCHING_CONFIG_FROM_NODE = M("BOOTH_FETCHING_CONFIG_FROM_NODE")
BOOTH_INVALID_NAME = M("BOOTH_INVALID_NAME")
BOOTH_LACK_OF_SITES = M("BOOTH_LACK_OF_SITES")
BOOTH_MULTIPLE_TIMES_IN_CIB = M("BOOTH_MULTIPLE_TIMES_IN_CIB")
BOOTH_NOT_EXISTS_IN_CIB = M("BOOTH_NOT_EXISTS_IN_CIB")
BOOTH_PATH_NOT_EXISTS = M("BOOTH_PATH_NOT_EXISTS")
BOOTH_PEERS_STATUS_ERROR = M("BOOTH_PEERS_STATUS_ERROR")
BOOTH_TICKET_DOES_NOT_EXIST = M("BOOTH_TICKET_DOES_NOT_EXIST")
BOOTH_TICKET_DUPLICATE = M("BOOTH_TICKET_DUPLICATE")
BOOTH_TICKET_NAME_INVALID = M("BOOTH_TICKET_NAME_INVALID")
BOOTH_TICKET_OPERATION_FAILED = M("BOOTH_TICKET_OPERATION_FAILED")
BOOTH_TICKET_STATUS_ERROR = M("BOOTH_TICKET_STATUS_ERROR")
BOOTH_UNSUPPORTED_FILE_LOCATION = M("BOOTH_UNSUPPORTED_FILE_LOCATION")
CANNOT_BAN_RESOURCE_MASTER_RESOURCE_NOT_PROMOTABLE = M(
    "CANNOT_BAN_RESOURCE_MASTER_RESOURCE_NOT_PROMOTABLE"
)
CANNOT_BAN_RESOURCE_STOPPED_NO_NODE_SPECIFIED = M(
    "CANNOT_BAN_RESOURCE_STOPPED_NO_NODE_SPECIFIED"
)
CANNOT_CREATE_DEFAULT_CLUSTER_PROPERTY_SET = M(
    "CANNOT_CREATE_DEFAULT_CLUSTER_PROPERTY_SET"
)
CANNOT_GROUP_RESOURCE_WRONG_TYPE = M("CANNOT_GROUP_RESOURCE_WRONG_TYPE")
CANNOT_LEAVE_GROUP_EMPTY_AFTER_MOVE = M("CANNOT_LEAVE_GROUP_EMPTY_AFTER_MOVE")
CANNOT_MOVE_RESOURCE_BUNDLE = M("CANNOT_MOVE_RESOURCE_BUNDLE")
CANNOT_MOVE_RESOURCE_CLONE = M("CANNOT_MOVE_RESOURCE_CLONE")
CANNOT_MOVE_RESOURCE_MASTER_RESOURCE_NOT_PROMOTABLE = M(
    "CANNOT_MOVE_RESOURCE_MASTER_RESOURCE_NOT_PROMOTABLE"
)
CANNOT_MOVE_RESOURCE_PROMOTABLE_INNER = M(
    "CANNOT_MOVE_RESOURCE_PROMOTABLE_INNER"
)
CANNOT_MOVE_RESOURCE_NOT_RUNNING = M("CANNOT_MOVE_RESOURCE_NOT_RUNNING")
CANNOT_MOVE_RESOURCE_STOPPED_NO_NODE_SPECIFIED = M(
    "CANNOT_MOVE_RESOURCE_STOPPED_NO_NODE_SPECIFIED"
)
CANNOT_REMOVE_ALL_CLUSTER_NODES = M("CANNOT_REMOVE_ALL_CLUSTER_NODES")
CANNOT_SET_ORDER_CONSTRAINTS_FOR_RESOURCES_IN_THE_SAME_GROUP = M(
    "CANNOT_SET_ORDER_CONSTRAINTS_FOR_RESOURCES_IN_THE_SAME_GROUP"
)
CANNOT_UNMOVE_UNBAN_RESOURCE_MASTER_RESOURCE_NOT_PROMOTABLE = M(
    "CANNOT_UNMOVE_UNBAN_RESOURCE_MASTER_RESOURCE_NOT_PROMOTABLE"
)
CIB_ACL_ROLE_IS_ALREADY_ASSIGNED_TO_TARGET = M(
    "CIB_ACL_ROLE_IS_ALREADY_ASSIGNED_TO_TARGET"
)
CIB_ACL_ROLE_IS_NOT_ASSIGNED_TO_TARGET = M(
    "CIB_ACL_ROLE_IS_NOT_ASSIGNED_TO_TARGET"
)
CIB_ACL_TARGET_ALREADY_EXISTS = M("CIB_ACL_TARGET_ALREADY_EXISTS")
CIB_ALERT_RECIPIENT_ALREADY_EXISTS = M("CIB_ALERT_RECIPIENT_ALREADY_EXISTS")
CIB_ALERT_RECIPIENT_VALUE_INVALID = M("CIB_ALERT_RECIPIENT_VALUE_INVALID")
CIB_CANNOT_FIND_MANDATORY_SECTION = M("CIB_CANNOT_FIND_MANDATORY_SECTION")
CIB_DIFF_ERROR = M("CIB_DIFF_ERROR")
CIB_FENCING_LEVEL_ALREADY_EXISTS = M("CIB_FENCING_LEVEL_ALREADY_EXISTS")
CIB_FENCING_LEVEL_DOES_NOT_EXIST = M("CIB_FENCING_LEVEL_DOES_NOT_EXIST")
CIB_LOAD_ERROR_BAD_FORMAT = M("CIB_LOAD_ERROR_BAD_FORMAT")
CIB_LOAD_ERROR = M("CIB_LOAD_ERROR")
CIB_LOAD_ERROR_GET_NODES_FOR_VALIDATION = M(
    "CIB_LOAD_ERROR_GET_NODES_FOR_VALIDATION"
)
CIB_NVSET_AMBIGUOUS_PROVIDE_NVSET_ID = M("CIB_NVSET_AMBIGUOUS_PROVIDE_NVSET_ID")
CIB_LOAD_ERROR_SCOPE_MISSING = M("CIB_LOAD_ERROR_SCOPE_MISSING")
CIB_PUSH_ERROR = M("CIB_PUSH_ERROR")
CIB_SAVE_TMP_ERROR = M("CIB_SAVE_TMP_ERROR")
CIB_SIMULATE_ERROR = M("CIB_SIMULATE_ERROR")
CIB_UPGRADE_FAILED = M("CIB_UPGRADE_FAILED")
CIB_UPGRADE_FAILED_TO_MINIMAL_REQUIRED_VERSION = M(
    "CIB_UPGRADE_FAILED_TO_MINIMAL_REQUIRED_VERSION"
)
CIB_UPGRADE_SUCCESSFUL = M("CIB_UPGRADE_SUCCESSFUL")
CLUSTER_DESTROY_STARTED = M("CLUSTER_DESTROY_STARTED")
CLUSTER_DESTROY_SUCCESS = M("CLUSTER_DESTROY_SUCCESS")
CLUSTER_ENABLE_STARTED = M("CLUSTER_ENABLE_STARTED")
CLUSTER_ENABLE_SUCCESS = M("CLUSTER_ENABLE_SUCCESS")
CLUSTER_RESTART_REQUIRED_TO_APPLY_CHANGES = M(
    "CLUSTER_RESTART_REQUIRED_TO_APPLY_CHANGES"
)
CLUSTER_SETUP_SUCCESS = M("CLUSTER_SETUP_SUCCESS")
CLUSTER_START_STARTED = M("CLUSTER_START_STARTED")
CLUSTER_START_SUCCESS = M("CLUSTER_START_SUCCESS")
CLUSTER_UUID_ALREADY_SET = M("CLUSTER_UUID_ALREADY_SET")
CLUSTER_WILL_BE_DESTROYED = M("CLUSTER_WILL_BE_DESTROYED")
COMMAND_INVALID_PAYLOAD = M("COMMAND_INVALID_PAYLOAD")
COMMAND_UNKNOWN = M("COMMAND_UNKNOWN")
LIVE_ENVIRONMENT_NOT_CONSISTENT = M("LIVE_ENVIRONMENT_NOT_CONSISTENT")
LIVE_ENVIRONMENT_REQUIRED = M("LIVE_ENVIRONMENT_REQUIRED")
LIVE_ENVIRONMENT_REQUIRED_FOR_LOCAL_NODE = M(
    "LIVE_ENVIRONMENT_REQUIRED_FOR_LOCAL_NODE"
)
RESOURCE_STONITH_COMMANDS_MISMATCH = M("RESOURCE_STONITH_COMMANDS_MISMATCH")
COROSYNC_ADDRESS_IP_VERSION_WRONG_FOR_LINK = M(
    "COROSYNC_ADDRESS_IP_VERSION_WRONG_FOR_LINK"
)
COROSYNC_AUTHKEY_WRONG_LENGTH = M("COROSYNC_AUTHKEY_WRONG_LENGTH")
COROSYNC_BAD_NODE_ADDRESSES_COUNT = M("COROSYNC_BAD_NODE_ADDRESSES_COUNT")
COROSYNC_CLUSTER_NAME_INVALID_FOR_GFS2 = M(
    "COROSYNC_CLUSTER_NAME_INVALID_FOR_GFS2"
)
COROSYNC_CONFIG_ACCEPTED_BY_NODE = M("COROSYNC_CONFIG_ACCEPTED_BY_NODE")
COROSYNC_CONFIG_CANNOT_SAVE_INVALID_NAMES_VALUES = M(
    "COROSYNC_CONFIG_CANNOT_SAVE_INVALID_NAMES_VALUES"
)
COROSYNC_CONFIG_DISTRIBUTION_NODE_ERROR = M(
    "COROSYNC_CONFIG_DISTRIBUTION_NODE_ERROR"
)
COROSYNC_CONFIG_DISTRIBUTION_STARTED = M("COROSYNC_CONFIG_DISTRIBUTION_STARTED")
COROSYNC_CONFIG_MISSING_IDS_OF_NODES = M("COROSYNC_CONFIG_MISSING_IDS_OF_NODES")
COROSYNC_CONFIG_MISSING_NAMES_OF_NODES = M(
    "COROSYNC_CONFIG_MISSING_NAMES_OF_NODES"
)
COROSYNC_CONFIG_NO_NODES_DEFINED = M("COROSYNC_CONFIG_NO_NODES_DEFINED")
COROSYNC_CONFIG_RELOADED = M("COROSYNC_CONFIG_RELOADED")
COROSYNC_CONFIG_RELOAD_ERROR = M("COROSYNC_CONFIG_RELOAD_ERROR")
COROSYNC_CONFIG_RELOAD_NOT_POSSIBLE = M("COROSYNC_CONFIG_RELOAD_NOT_POSSIBLE")
COROSYNC_CONFIG_UNSUPPORTED_TRANSPORT = M(
    "COROSYNC_CONFIG_UNSUPPORTED_TRANSPORT"
)
COROSYNC_IP_VERSION_MISMATCH_IN_LINKS = M(
    "COROSYNC_IP_VERSION_MISMATCH_IN_LINKS"
)
COROSYNC_CANNOT_ADD_REMOVE_LINKS_BAD_TRANSPORT = M(
    "COROSYNC_CANNOT_ADD_REMOVE_LINKS_BAD_TRANSPORT"
)
COROSYNC_CANNOT_ADD_REMOVE_LINKS_NO_LINKS_SPECIFIED = M(
    "COROSYNC_CANNOT_ADD_REMOVE_LINKS_NO_LINKS_SPECIFIED"
)
COROSYNC_CANNOT_ADD_REMOVE_LINKS_TOO_MANY_FEW_LINKS = M(
    "COROSYNC_CANNOT_ADD_REMOVE_LINKS_TOO_MANY_FEW_LINKS"
)
COROSYNC_LINK_ALREADY_EXISTS_CANNOT_ADD = M(
    "COROSYNC_LINK_ALREADY_EXISTS_CANNOT_ADD"
)
COROSYNC_LINK_DOES_NOT_EXIST_CANNOT_REMOVE = M(
    "COROSYNC_LINK_DOES_NOT_EXIST_CANNOT_REMOVE"
)
COROSYNC_LINK_DOES_NOT_EXIST_CANNOT_UPDATE = M(
    "COROSYNC_LINK_DOES_NOT_EXIST_CANNOT_UPDATE"
)
COROSYNC_LINK_NUMBER_DUPLICATION = M("COROSYNC_LINK_NUMBER_DUPLICATION")
COROSYNC_NODE_ADDRESS_COUNT_MISMATCH = M("COROSYNC_NODE_ADDRESS_COUNT_MISMATCH")
COROSYNC_NODE_CONFLICT_CHECK_SKIPPED = M("COROSYNC_NODE_CONFLICT_CHECK_SKIPPED")
COROSYNC_NODES_MISSING = M("COROSYNC_NODES_MISSING")
COROSYNC_NOT_RUNNING_CHECK_NODE_ERROR = M(
    "COROSYNC_NOT_RUNNING_CHECK_NODE_ERROR"
)
COROSYNC_NOT_RUNNING_CHECK_STARTED = M("COROSYNC_NOT_RUNNING_CHECK_STARTED")
COROSYNC_NOT_RUNNING_ON_NODE = M("COROSYNC_NOT_RUNNING_ON_NODE")
COROSYNC_OPTIONS_INCOMPATIBLE_WITH_QDEVICE = M(
    "COROSYNC_OPTIONS_INCOMPATIBLE_WITH_QDEVICE"
)
COROSYNC_QUORUM_ATB_CANNOT_BE_DISABLED_DUE_TO_SBD = M(
    "COROSYNC_QUORUM_ATB_CANNOT_BE_DISABLED_DUE_TO_SBD"
)
COROSYNC_QUORUM_ATB_WILL_BE_ENABLED_DUE_TO_SBD = M(
    "COROSYNC_QUORUM_ATB_WILL_BE_ENABLED_DUE_TO_SBD"
)
COROSYNC_QUORUM_GET_STATUS_ERROR = M("COROSYNC_QUORUM_GET_STATUS_ERROR")
COROSYNC_QUORUM_HEURISTICS_ENABLED_WITH_NO_EXEC = M(
    "COROSYNC_QUORUM_HEURISTICS_ENABLED_WITH_NO_EXEC"
)
COROSYNC_QUORUM_LOSS_UNABLE_TO_CHECK = M("COROSYNC_QUORUM_LOSS_UNABLE_TO_CHECK")
COROSYNC_QUORUM_SET_EXPECTED_VOTES_ERROR = M(
    "COROSYNC_QUORUM_SET_EXPECTED_VOTES_ERROR"
)
COROSYNC_QUORUM_WILL_BE_LOST = M("COROSYNC_QUORUM_WILL_BE_LOST")
COROSYNC_RUNNING_ON_NODE = M("COROSYNC_RUNNING_ON_NODE")
COROSYNC_TOO_MANY_LINKS_OPTIONS = M("COROSYNC_TOO_MANY_LINKS_OPTIONS")
COROSYNC_TRANSPORT_UNSUPPORTED_OPTIONS = M(
    "COROSYNC_TRANSPORT_UNSUPPORTED_OPTIONS"
)
CRM_MON_ERROR = M("CRM_MON_ERROR")
DEFAULTS_CAN_BE_OVERRIDDEN = M("DEFAULTS_CAN_BE_OVERRIDDEN")
DEPRECATED_OPTION = M("DEPRECATED_OPTION")
DEPRECATED_OPTION_VALUE = M("DEPRECATED_OPTION_VALUE")
DR_CONFIG_ALREADY_EXIST = M("DR_CONFIG_ALREADY_EXIST")
DR_CONFIG_DOES_NOT_EXIST = M("DR_CONFIG_DOES_NOT_EXIST")
DUPLICATE_CONSTRAINTS_EXIST = M("DUPLICATE_CONSTRAINTS_EXIST")
DUPLICATE_CONSTRAINTS_LIST = M("DUPLICATE_CONSTRAINTS_LIST")
EMPTY_RESOURCE_SET_LIST = M("EMPTY_RESOURCE_SET_LIST")
FENCE_HISTORY_COMMAND_ERROR = M("FENCE_HISTORY_COMMAND_ERROR")
FENCE_HISTORY_NOT_SUPPORTED = M("FENCE_HISTORY_NOT_SUPPORTED")
FILES_DISTRIBUTION_SKIPPED = M("FILES_DISTRIBUTION_SKIPPED")
FILES_DISTRIBUTION_STARTED = M("FILES_DISTRIBUTION_STARTED")
FILES_REMOVE_FROM_NODES_STARTED = M("FILES_REMOVE_FROM_NODES_STARTED")
FILES_REMOVE_FROM_NODES_SKIPPED = M("FILES_REMOVE_FROM_NODES_SKIPPED")
FILE_ALREADY_EXISTS = M("FILE_ALREADY_EXISTS")
FILE_DISTRIBUTION_ERROR = M("FILE_DISTRIBUTION_ERROR")
FILE_DISTRIBUTION_SUCCESS = M("FILE_DISTRIBUTION_SUCCESS")
FILE_IO_ERROR = M("FILE_IO_ERROR")
FILE_REMOVE_FROM_NODE_ERROR = M("FILE_REMOVE_FROM_NODE_ERROR")
FILE_REMOVE_FROM_NODE_SUCCESS = M("FILE_REMOVE_FROM_NODE_SUCCESS")
HOST_NOT_FOUND = M("HOST_NOT_FOUND")
HOST_ALREADY_AUTHORIZED = M("HOST_ALREADY_AUTHORIZED")
HOST_ALREADY_IN_CLUSTER_CONFIG = M("HOST_ALREADY_IN_CLUSTER_CONFIG")
HOST_ALREADY_IN_CLUSTER_SERVICES = M("HOST_ALREADY_IN_CLUSTER_SERVICES")
ID_ALREADY_EXISTS = M("ID_ALREADY_EXISTS")
ID_BELONGS_TO_UNEXPECTED_TYPE = M("ID_BELONGS_TO_UNEXPECTED_TYPE")
ID_NOT_FOUND = M("ID_NOT_FOUND")
INVALID_CIB_CONTENT = M("INVALID_CIB_CONTENT")
INVALID_ID_BAD_CHAR = M("INVALID_ID_BAD_CHAR")
INVALID_ID_IS_EMPTY = M("INVALID_ID_IS_EMPTY")
INVALID_OPTIONS = M("INVALID_OPTIONS")
INVALID_USERDEFINED_OPTIONS = M("INVALID_USERDEFINED_OPTIONS")
INVALID_OPTION_TYPE = M("INVALID_OPTION_TYPE")
INVALID_OPTION_VALUE = M("INVALID_OPTION_VALUE")
INVALID_RESOURCE_AGENT_NAME = M("INVALID_RESOURCE_AGENT_NAME")
INVALID_RESPONSE_FORMAT = M("INVALID_RESPONSE_FORMAT")
INVALID_SCORE = M("INVALID_SCORE")
INVALID_STONITH_AGENT_NAME = M("INVALID_STONITH_AGENT_NAME")
INVALID_TIMEOUT_VALUE = M("INVALID_TIMEOUT_VALUE")
MULTIPLE_RESULTS_FOUND = M("MULTIPLE_RESULTS_FOUND")
MUTUALLY_EXCLUSIVE_OPTIONS = M("MUTUALLY_EXCLUSIVE_OPTIONS")
NO_ACTION_NECESSARY = M("NO_ACTION_NECESSARY")
NODE_ADDRESSES_ALREADY_EXIST = M("NODE_ADDRESSES_ALREADY_EXIST")
NODE_ADDRESSES_CANNOT_BE_EMPTY = M("NODE_ADDRESSES_CANNOT_BE_EMPTY")
NODE_ADDRESSES_DUPLICATION = M("NODE_ADDRESSES_DUPLICATION")
NODE_ADDRESSES_UNRESOLVABLE = M("NODE_ADDRESSES_UNRESOLVABLE")
NODE_COMMUNICATION_COMMAND_UNSUCCESSFUL = M(
    "NODE_COMMUNICATION_COMMAND_UNSUCCESSFUL"
)
NODE_COMMUNICATION_DEBUG_INFO = M("NODE_COMMUNICATION_DEBUG_INFO")
NODE_COMMUNICATION_ERROR = M("NODE_COMMUNICATION_ERROR")
NODE_COMMUNICATION_ERROR_NOT_AUTHORIZED = M(
    "NODE_COMMUNICATION_ERROR_NOT_AUTHORIZED"
)
NODE_COMMUNICATION_ERROR_PERMISSION_DENIED = M(
    "NODE_COMMUNICATION_ERROR_PERMISSION_DENIED"
)
NODE_COMMUNICATION_ERROR_UNABLE_TO_CONNECT = M(
    "NODE_COMMUNICATION_ERROR_UNABLE_TO_CONNECT"
)
NODE_COMMUNICATION_ERROR_UNSUPPORTED_COMMAND = M(
    "NODE_COMMUNICATION_ERROR_UNSUPPORTED_COMMAND"
)
NODE_COMMUNICATION_ERROR_TIMED_OUT = M("NODE_COMMUNICATION_ERROR_TIMED_OUT")
NODE_COMMUNICATION_FINISHED = M("NODE_COMMUNICATION_FINISHED")
NODE_COMMUNICATION_NOT_CONNECTED = M("NODE_COMMUNICATION_NOT_CONNECTED")
NODE_COMMUNICATION_NO_MORE_ADDRESSES = M("NODE_COMMUNICATION_NO_MORE_ADDRESSES")
NODE_COMMUNICATION_PROXY_IS_SET = M("NODE_COMMUNICATION_PROXY_IS_SET")
NODE_COMMUNICATION_RETRYING = M("NODE_COMMUNICATION_RETRYING")
NODE_COMMUNICATION_STARTED = M("NODE_COMMUNICATION_STARTED")
NODE_NAMES_ALREADY_EXIST = M("NODE_NAMES_ALREADY_EXIST")
NODE_NAMES_DUPLICATION = M("NODE_NAMES_DUPLICATION")
NODE_NOT_FOUND = M("NODE_NOT_FOUND")
NODE_REMOVE_IN_PACEMAKER_FAILED = M("NODE_REMOVE_IN_PACEMAKER_FAILED")
NONE_HOST_FOUND = M("NONE_HOST_FOUND")
NODE_USED_AS_TIE_BREAKER = M("NODE_USED_AS_TIE_BREAKER")
NODES_TO_REMOVE_UNREACHABLE = M("NODES_TO_REMOVE_UNREACHABLE")
NODE_TO_CLEAR_IS_STILL_IN_CLUSTER = M("NODE_TO_CLEAR_IS_STILL_IN_CLUSTER")
NODE_IN_LOCAL_CLUSTER = M("NODE_IN_LOCAL_CLUSTER")
NOT_AUTHORIZED = M("NOT_AUTHORIZED")
OBJECT_WITH_ID_IN_UNEXPECTED_CONTEXT = M("OBJECT_WITH_ID_IN_UNEXPECTED_CONTEXT")
OMITTING_NODE = M("OMITTING_NODE")
PACEMAKER_SIMULATION_RESULT = M("PACEMAKER_SIMULATION_RESULT")
PACEMAKER_LOCAL_NODE_NAME_NOT_FOUND = M("PACEMAKER_LOCAL_NODE_NAME_NOT_FOUND")
PARSE_ERROR_COROSYNC_CONF = M("PARSE_ERROR_COROSYNC_CONF")
PARSE_ERROR_COROSYNC_CONF_EXTRA_CHARACTERS_AFTER_OPENING_BRACE = M(
    "PARSE_ERROR_COROSYNC_CONF_EXTRA_CHARACTERS_AFTER_OPENING_BRACE"
)
PARSE_ERROR_COROSYNC_CONF_EXTRA_CHARACTERS_BEFORE_OR_AFTER_CLOSING_BRACE = M(
    "PARSE_ERROR_COROSYNC_CONF_EXTRA_CHARACTERS_BEFORE_OR_AFTER_CLOSING_BRACE"
)
PARSE_ERROR_COROSYNC_CONF_LINE_IS_NOT_SECTION_NOR_KEY_VALUE = M(
    "PARSE_ERROR_COROSYNC_CONF_LINE_IS_NOT_SECTION_NOR_KEY_VALUE"
)
PARSE_ERROR_COROSYNC_CONF_MISSING_CLOSING_BRACE = M(
    "PARSE_ERROR_COROSYNC_CONF_MISSING_CLOSING_BRACE"
)
PARSE_ERROR_COROSYNC_CONF_MISSING_SECTION_NAME_BEFORE_OPENING_BRACE = M(
    "PARSE_ERROR_COROSYNC_CONF_MISSING_SECTION_NAME_BEFORE_OPENING_BRACE"
)
PARSE_ERROR_COROSYNC_CONF_UNEXPECTED_CLOSING_BRACE = M(
    "PARSE_ERROR_COROSYNC_CONF_UNEXPECTED_CLOSING_BRACE"
)
PARSE_ERROR_JSON_FILE = M("PARSE_ERROR_JSON_FILE")
PCSD_VERSION_TOO_OLD = M("PCSD_VERSION_TOO_OLD")
PCSD_SSL_CERT_AND_KEY_DISTRIBUTION_STARTED = M(
    "PCSD_SSL_CERT_AND_KEY_DISTRIBUTION_STARTED"
)
PCSD_SSL_CERT_AND_KEY_SET_SUCCESS = M("PCSD_SSL_CERT_AND_KEY_SET_SUCCESS")
PREREQUISITE_OPTION_MUST_BE_ENABLED_AS_WELL = M(
    "PREREQUISITE_OPTION_MUST_BE_ENABLED_AS_WELL"
)
PREREQUISITE_OPTION_MUST_BE_DISABLED = M("PREREQUISITE_OPTION_MUST_BE_DISABLED")
PREREQUISITE_OPTION_MUST_NOT_BE_SET = M("PREREQUISITE_OPTION_MUST_NOT_BE_SET")
PREREQUISITE_OPTION_IS_MISSING = M("PREREQUISITE_OPTION_IS_MISSING")
QDEVICE_ALREADY_DEFINED = M("QDEVICE_ALREADY_DEFINED")
QDEVICE_ALREADY_INITIALIZED = M("QDEVICE_ALREADY_INITIALIZED")
QDEVICE_CERTIFICATE_ACCEPTED_BY_NODE = M("QDEVICE_CERTIFICATE_ACCEPTED_BY_NODE")
QDEVICE_CERTIFICATE_BAD_FORMAT = M("QDEVICE_CERTIFICATE_BAD_FORMAT")
QDEVICE_CERTIFICATE_DISTRIBUTION_STARTED = M(
    "QDEVICE_CERTIFICATE_DISTRIBUTION_STARTED"
)
QDEVICE_CERTIFICATE_IMPORT_ERROR = M("QDEVICE_CERTIFICATE_IMPORT_ERROR")
QDEVICE_CERTIFICATE_READ_ERROR = M("QDEVICE_CERTIFICATE_READ_ERROR")
QDEVICE_CERTIFICATE_REMOVAL_STARTED = M("QDEVICE_CERTIFICATE_REMOVAL_STARTED")
QDEVICE_CERTIFICATE_REMOVED_FROM_NODE = M(
    "QDEVICE_CERTIFICATE_REMOVED_FROM_NODE"
)
QDEVICE_CERTIFICATE_SIGN_ERROR = M("QDEVICE_CERTIFICATE_SIGN_ERROR")
QDEVICE_DESTROY_ERROR = M("QDEVICE_DESTROY_ERROR")
QDEVICE_DESTROY_SUCCESS = M("QDEVICE_DESTROY_SUCCESS")
QDEVICE_GET_STATUS_ERROR = M("QDEVICE_GET_STATUS_ERROR")
QDEVICE_INITIALIZATION_ERROR = M("QDEVICE_INITIALIZATION_ERROR")
QDEVICE_INITIALIZATION_SUCCESS = M("QDEVICE_INITIALIZATION_SUCCESS")
QDEVICE_NOT_DEFINED = M("QDEVICE_NOT_DEFINED")
QDEVICE_NOT_INITIALIZED = M("QDEVICE_NOT_INITIALIZED")
QDEVICE_NOT_RUNNING = M("QDEVICE_NOT_RUNNING")
QDEVICE_CLIENT_RELOAD_STARTED = M("QDEVICE_CLIENT_RELOAD_STARTED")
QDEVICE_REMOVE_OR_CLUSTER_STOP_NEEDED = M(
    "QDEVICE_REMOVE_OR_CLUSTER_STOP_NEEDED"
)
QDEVICE_USED_BY_CLUSTERS = M("QDEVICE_USED_BY_CLUSTERS")
CLONING_STONITH_RESOURCES_HAS_NO_EFFECT = M(
    "CLONING_STONITH_RESOURCES_HAS_NO_EFFECT"
)
REQUIRED_OPTIONS_ARE_MISSING = M("REQUIRED_OPTIONS_ARE_MISSING")
REQUIRED_OPTION_OF_ALTERNATIVES_IS_MISSING = M(
    "REQUIRED_OPTION_OF_ALTERNATIVES_IS_MISSING"
)
RESOURCE_BAN_PCMK_ERROR = M("RESOURCE_BAN_PCMK_ERROR")
RESOURCE_BAN_PCMK_SUCCESS = M("RESOURCE_BAN_PCMK_SUCCESS")
RESOURCE_BUNDLE_ALREADY_CONTAINS_A_RESOURCE = M(
    "RESOURCE_BUNDLE_ALREADY_CONTAINS_A_RESOURCE"
)
RESOURCE_BUNDLE_UNSUPPORTED_CONTAINER_TYPE = M(
    "RESOURCE_BUNDLE_UNSUPPORTED_CONTAINER_TYPE"
)
RESOURCE_CLEANUP_ERROR = M("RESOURCE_CLEANUP_ERROR")
RESOURCE_CLONE_INCOMPATIBLE_META_ATTRIBUTES = M(
    "RESOURCE_CLONE_INCOMPATIBLE_META_ATTRIBUTES"
)
RESOURCE_DOES_NOT_RUN = M("RESOURCE_DOES_NOT_RUN")
RESOURCE_DISABLE_AFFECTS_OTHER_RESOURCES = M(
    "RESOURCE_DISABLE_AFFECTS_OTHER_RESOURCES"
)
RESOURCE_FOR_CONSTRAINT_IS_MULTIINSTANCE = M(
    "RESOURCE_FOR_CONSTRAINT_IS_MULTIINSTANCE"
)
RESOURCE_IN_BUNDLE_NOT_ACCESSIBLE = M("RESOURCE_IN_BUNDLE_NOT_ACCESSIBLE")
RESOURCE_INSTANCE_ATTR_VALUE_NOT_UNIQUE = M(
    "RESOURCE_INSTANCE_ATTR_VALUE_NOT_UNIQUE"
)
RESOURCE_INSTANCE_ATTR_GROUP_VALUE_NOT_UNIQUE = M(
    "RESOURCE_INSTANCE_ATTR_GROUP_VALUE_NOT_UNIQUE"
)
RESOURCE_IS_GUEST_NODE_ALREADY = M("RESOURCE_IS_GUEST_NODE_ALREADY")
RESOURCE_IS_UNMANAGED = M("RESOURCE_IS_UNMANAGED")
RESOURCE_MANAGED_NO_MONITOR_ENABLED = M("RESOURCE_MANAGED_NO_MONITOR_ENABLED")
RESOURCE_MOVE_PCMK_ERROR = M("RESOURCE_MOVE_PCMK_ERROR")
RESOURCE_MOVE_PCMK_SUCCESS = M("RESOURCE_MOVE_PCMK_SUCCESS")
RESOURCE_OPERATION_INTERVAL_ADAPTED = M("RESOURCE_OPERATION_INTERVAL_ADAPTED")
RESOURCE_OPERATION_INTERVAL_DUPLICATION = M(
    "RESOURCE_OPERATION_INTERVAL_DUPLICATION"
)
RESOURCE_REFRESH_ERROR = M("RESOURCE_REFRESH_ERROR")
RESOURCE_REFRESH_TOO_TIME_CONSUMING = M("RESOURCE_REFRESH_TOO_TIME_CONSUMING")
RESOURCE_RUNNING_ON_NODES = M("RESOURCE_RUNNING_ON_NODES")
RESOURCE_UNMOVE_UNBAN_PCMK_ERROR = M("RESOURCE_UNMOVE_UNBAN_PCMK_ERROR")
RESOURCE_UNMOVE_UNBAN_PCMK_SUCCESS = M("RESOURCE_UNMOVE_UNBAN_PCMK_SUCCESS")
RESOURCE_UNMOVE_UNBAN_PCMK_EXPIRED_NOT_SUPPORTED = M(
    "RESOURCE_UNMOVE_UNBAN_PCMK_EXPIRED_NOT_SUPPORTED"
)
RESOURCE_MAY_OR_MAY_NOT_MOVE = M("RESOURCE_MAY_OR_MAY_NOT_MOVE")
RESOURCE_MOVE_CONSTRAINT_CREATED = M("RESOURCE_MOVE_CONSTRAINT_CREATED")
RESOURCE_MOVE_CONSTRAINT_REMOVED = M("RESOURCE_MOVE_CONSTRAINT_REMOVED")
RESOURCE_MOVE_NOT_AFFECTING_RESOURCE = M("RESOURCE_MOVE_NOT_AFFECTING_RESOURCE")
RESOURCE_MOVE_AFFECTS_OTRHER_RESOURCES = M(
    "RESOURCE_MOVE_AFFECTS_OTRHER_RESOURCES"
)
RESOURCE_MOVE_AUTOCLEAN_SIMULATION_FAILURE = M(
    "RESOURCE_MOVE_AUTOCLEAN_SIMULATION_FAILURE"
)
RULE_IN_EFFECT_STATUS_DETECTION_NOT_SUPPORTED = M(
    "RULE_IN_EFFECT_STATUS_DETECTION_NOT_SUPPORTED"
)
RULE_EXPRESSION_NOT_ALLOWED = M("RULE_EXPRESSION_NOT_ALLOWED")
RULE_EXPRESSION_OPTIONS_DUPLICATION = M("RULE_EXPRESSION_OPTIONS_DUPLICATION")
RULE_EXPRESSION_PARSE_ERROR = M("RULE_EXPRESSION_PARSE_ERROR")
RULE_EXPRESSION_SINCE_GREATER_THAN_UNTIL = M(
    "RULE_EXPRESSION_SINCE_GREATER_THAN_UNTIL"
)
RUN_EXTERNAL_PROCESS_ERROR = M("RUN_EXTERNAL_PROCESS_ERROR")
RUN_EXTERNAL_PROCESS_FINISHED = M("RUN_EXTERNAL_PROCESS_FINISHED")
RUN_EXTERNAL_PROCESS_STARTED = M("RUN_EXTERNAL_PROCESS_STARTED")
SBD_CHECK_STARTED = M("SBD_CHECK_STARTED")
SBD_CHECK_SUCCESS = M("SBD_CHECK_SUCCESS")
SBD_CONFIG_ACCEPTED_BY_NODE = M("SBD_CONFIG_ACCEPTED_BY_NODE")
SBD_CONFIG_DISTRIBUTION_STARTED = M("SBD_CONFIG_DISTRIBUTION_STARTED")
SBD_DEVICE_DOES_NOT_EXIST = M("SBD_DEVICE_DOES_NOT_EXIST")
SBD_DEVICE_DUMP_ERROR = M("SBD_DEVICE_DUMP_ERROR")
SBD_DEVICE_INITIALIZATION_ERROR = M("SBD_DEVICE_INITIALIZATION_ERROR")
SBD_DEVICE_INITIALIZATION_STARTED = M("SBD_DEVICE_INITIALIZATION_STARTED")
SBD_DEVICE_INITIALIZATION_SUCCESS = M("SBD_DEVICE_INITIALIZATION_SUCCESS")
SBD_DEVICE_IS_NOT_BLOCK_DEVICE = M("SBD_DEVICE_IS_NOT_BLOCK_DEVICE")
SBD_DEVICE_LIST_ERROR = M("SBD_DEVICE_LIST_ERROR")
SBD_DEVICE_MESSAGE_ERROR = M("SBD_DEVICE_MESSAGE_ERROR")
SBD_DEVICE_PATH_NOT_ABSOLUTE = M("SBD_DEVICE_PATH_NOT_ABSOLUTE")
SBD_LIST_WATCHDOG_ERROR = M("SBD_LIST_WATCHDOG_ERROR")
SBD_NO_DEVICE_FOR_NODE = M("SBD_NO_DEVICE_FOR_NODE")
SBD_NOT_USED_CANNOT_SET_SBD_OPTIONS = M("SBD_NOT_USED_CANNOT_SET_SBD_OPTIONS")
SBD_TOO_MANY_DEVICES_FOR_NODE = M("SBD_TOO_MANY_DEVICES_FOR_NODE")
SBD_WITH_DEVICES_NOT_USED_CANNOT_SET_DEVICE = M(
    "SBD_WITH_DEVICES_NOT_USED_CANNOT_SET_DEVICE"
)
SBD_WATCHDOG_NOT_SUPPORTED = M("SBD_WATCHDOG_NOT_SUPPORTED")
SBD_WATCHDOG_VALIDATION_INACTIVE = M("SBD_WATCHDOG_VALIDATION_INACTIVE")
SBD_WATCHDOG_TEST_ERROR = M("SBD_WATCHDOG_TEST_ERROR")
SBD_WATCHDOG_TEST_MULTIPLE_DEVICES = M("SBD_WATCHDOG_TEST_MULTIPLE_DEVICES")
SBD_WATCHDOG_TEST_FAILED = M("SBD_WATCHDOG_TEST_FAILED")
SERVICE_ACTION_STARTED = M("SERVICE_ACTION_STARTED")
SERVICE_ACTION_FAILED = M("SERVICE_ACTION_FAILED")
SERVICE_ACTION_SUCCEEDED = M("SERVICE_ACTION_SUCCEEDED")
SERVICE_ACTION_SKIPPED = M("SERVICE_ACTION_SKIPPED")
SERVICE_NOT_INSTALLED = M("SERVICE_NOT_INSTALLED")
SERVICE_VERSION_MISMATCH = M("SERVICE_VERSION_MISMATCH")
UNABLE_TO_GET_RESOURCE_OPERATION_DIGESTS = M(
    "UNABLE_TO_GET_RESOURCE_OPERATION_DIGESTS"
)
STONITH_WATCHDOG_TIMEOUT_CANNOT_BE_SET = M(
    "STONITH_WATCHDOG_TIMEOUT_CANNOT_BE_SET"
)
STONITH_WATCHDOG_TIMEOUT_CANNOT_BE_UNSET = M(
    "STONITH_WATCHDOG_TIMEOUT_CANNOT_BE_UNSET"
)
STONITH_WATCHDOG_TIMEOUT_TOO_SMALL = M("STONITH_WATCHDOG_TIMEOUT_TOO_SMALL")
STONITH_RESOURCES_DO_NOT_EXIST = M("STONITH_RESOURCES_DO_NOT_EXIST")
STONITH_RESTARTLESS_UPDATE_MISSING_MPATH_KEYS = M(
    "STONITH_RESTARTLESS_UPDATE_MISSING_MPATH_KEYS"
)
STONITH_RESTARTLESS_UPDATE_OF_SCSI_DEVICES_NOT_SUPPORTED = M(
    "STONITH_RESTARTLESS_UPDATE_OF_SCSI_DEVICES_NOT_SUPPORTED"
)
STONITH_RESTARTLESS_UPDATE_UNSUPPORTED_AGENT = M(
    "STONITH_RESTARTLESS_UPDATE_UNSUPPORTED_AGENT"
)
STONITH_UNFENCING_FAILED = M("STONITH_UNFENCING_FAILED")
STONITH_UNFENCING_DEVICE_STATUS_FAILED = M(
    "STONITH_UNFENCING_DEVICE_STATUS_FAILED"
)
STONITH_UNFENCING_SKIPPED_DEVICES_FENCED = M(
    "STONITH_UNFENCING_SKIPPED_DEVICES_FENCED"
)
STONITH_RESTARTLESS_UPDATE_UNABLE_TO_PERFORM = M(
    "STONITH_RESTARTLESS_UPDATE_UNABLE_TO_PERFORM"
)
SERVICE_COMMANDS_ON_NODES_STARTED = M("SERVICE_COMMANDS_ON_NODES_STARTED")
SERVICE_COMMANDS_ON_NODES_SKIPPED = M("SERVICE_COMMANDS_ON_NODES_SKIPPED")
SERVICE_COMMAND_ON_NODE_ERROR = M("SERVICE_COMMAND_ON_NODE_ERROR")
SERVICE_COMMAND_ON_NODE_SUCCESS = M("SERVICE_COMMAND_ON_NODE_SUCCESS")
SERVICE_UNABLE_TO_DETECT_INIT_SYSTEM = M("SERVICE_UNABLE_TO_DETECT_INIT_SYSTEM")
SYSTEM_WILL_RESET = M("SYSTEM_WILL_RESET")
# TODO: remove, use ADD_REMOVE reports
TAG_ADD_REMOVE_IDS_DUPLICATION = M("TAG_ADD_REMOVE_IDS_DUPLICATION")
# TODO: remove, use ADD_REMOVE reports
TAG_ADJACENT_REFERENCE_ID_NOT_IN_THE_TAG = M(
    "TAG_ADJACENT_REFERENCE_ID_NOT_IN_THE_TAG"
)
# TODO: remove, use ADD_REMOVE reports
TAG_CANNOT_ADD_AND_REMOVE_IDS_AT_THE_SAME_TIME = M(
    "TAG_CANNOT_ADD_AND_REMOVE_IDS_AT_THE_SAME_TIME"
)
# TODO: remove, use ADD_REMOVE reports
TAG_CANNOT_ADD_REFERENCE_IDS_ALREADY_IN_THE_TAG = M(
    "TAG_CANNOT_ADD_REFERENCE_IDS_ALREADY_IN_THE_TAG"
)
TAG_CANNOT_CONTAIN_ITSELF = M("TAG_CANNOT_CONTAIN_ITSELF")
TAG_CANNOT_CREATE_EMPTY_TAG_NO_IDS_SPECIFIED = M(
    "TAG_CANNOT_CREATE_EMPTY_TAG_NO_IDS_SPECIFIED"
)
# TODO: remove, use ADD_REMOVE reports
TAG_CANNOT_PUT_ID_NEXT_TO_ITSELF = M("TAG_CANNOT_PUT_ID_NEXT_TO_ITSELF")
# TODO: remove, use ADD_REMOVE reports
TAG_CANNOT_REMOVE_ADJACENT_ID = M("TAG_CANNOT_REMOVE_ADJACENT_ID")
# TODO: remove, use ADD_REMOVE reports
TAG_CANNOT_REMOVE_REFERENCES_WITHOUT_REMOVING_TAG = M(
    "TAG_CANNOT_REMOVE_REFERENCES_WITHOUT_REMOVING_TAG"
)
TAG_CANNOT_REMOVE_TAG_REFERENCED_IN_CONSTRAINTS = M(
    "TAG_CANNOT_REMOVE_TAG_REFERENCED_IN_CONSTRAINTS"
)
TAG_CANNOT_REMOVE_TAGS_NO_TAGS_SPECIFIED = M(
    "TAG_CANNOT_REMOVE_TAGS_NO_TAGS_SPECIFIED"
)
# TODO: remove, use ADD_REMOVE reports
TAG_CANNOT_SPECIFY_ADJACENT_ID_WITHOUT_IDS_TO_ADD = M(
    "TAG_CANNOT_SPECIFY_ADJACENT_ID_WITHOUT_IDS_TO_ADD"
)
# TODO: remove, use ADD_REMOVE reports
TAG_CANNOT_UPDATE_TAG_NO_IDS_SPECIFIED = M(
    "TAG_CANNOT_UPDATE_TAG_NO_IDS_SPECIFIED"
)
# TODO: remove, use ADD_REMOVE reports
TAG_IDS_NOT_IN_THE_TAG = M("TAG_IDS_NOT_IN_THE_TAG")
TMP_FILE_WRITE = M("TMP_FILE_WRITE")
UNABLE_TO_CONNECT_TO_ANY_REMAINING_NODE = M(
    "UNABLE_TO_CONNECT_TO_ANY_REMAINING_NODE"
)
UNABLE_TO_CONNECT_TO_ALL_REMAINING_NODE = M(
    "UNABLE_TO_CONNECT_TO_ALL_REMAINING_NODE"
)
UNABLE_TO_GET_AGENT_METADATA = M("UNABLE_TO_GET_AGENT_METADATA")
UNABLE_TO_GET_SBD_CONFIG = M("UNABLE_TO_GET_SBD_CONFIG")
UNABLE_TO_GET_SBD_STATUS = M("UNABLE_TO_GET_SBD_STATUS")
UNABLE_TO_PERFORM_OPERATION_ON_ANY_NODE = M(
    "UNABLE_TO_PERFORM_OPERATION_ON_ANY_NODE"
)
WATCHDOG_INVALID = M("WATCHDOG_INVALID")
UNSUPPORTED_OPERATION_ON_NON_SYSTEMD_SYSTEMS = M(
    "UNSUPPORTED_OPERATION_ON_NON_SYSTEMD_SYSTEMS"
)
USE_COMMAND_NODE_ADD_REMOTE = M("USE_COMMAND_NODE_ADD_REMOTE")
USE_COMMAND_NODE_ADD_GUEST = M("USE_COMMAND_NODE_ADD_GUEST")
USE_COMMAND_NODE_REMOVE_GUEST = M("USE_COMMAND_NODE_REMOVE_GUEST")
USING_DEFAULT_ADDRESS_FOR_HOST = M("USING_DEFAULT_ADDRESS_FOR_HOST")
USING_DEFAULT_WATCHDOG = M("USING_DEFAULT_WATCHDOG")
WAIT_FOR_IDLE_STARTED = M("WAIT_FOR_IDLE_STARTED")
WAIT_FOR_IDLE_ERROR = M("WAIT_FOR_IDLE_ERROR")
WAIT_FOR_IDLE_NOT_LIVE_CLUSTER = M("WAIT_FOR_IDLE_NOT_LIVE_CLUSTER")
WAIT_FOR_IDLE_TIMED_OUT = M("WAIT_FOR_IDLE_TIMED_OUT")
WAIT_FOR_NODE_STARTUP_ERROR = M("WAIT_FOR_NODE_STARTUP_ERROR")
WAIT_FOR_NODE_STARTUP_STARTED = M("WAIT_FOR_NODE_STARTUP_STARTED")
WAIT_FOR_NODE_STARTUP_TIMED_OUT = M("WAIT_FOR_NODE_STARTUP_TIMED_OUT")
WAIT_FOR_NODE_STARTUP_WITHOUT_START = M("WAIT_FOR_NODE_STARTUP_WITHOUT_START")
WATCHDOG_NOT_FOUND = M("WATCHDOG_NOT_FOUND")
